"""
These tools help you identify, quantify, and visualize spatial patterns in your data.

calculate_density takes known quantities of some phenomenon and spreads these quantities across the map.
find_hot_spots identifies statistically significant clustering in the spatial pattern of your data.
"""
import json as _json
from datetime import datetime as _datetime
import logging as _logging
import arcgis as _arcgis
from arcgis.features import FeatureSet as _FeatureSet
from arcgis.features import Table as _Table
from arcgis.geoprocessing._support import _execute_gp_tool
from arcgis.geoprocessing import DataFile
from ._util import _id_generator, _feature_input, _set_context, _create_output_service, GAJob

_log=_logging.getLogger(__name__)

_use_async = True
#--------------------------------------------------------------------------
def forest(input_layer,
           var_prediction,
           var_explanatory,
           trees,
           max_tree_depth=None,
           random_vars=None,
           sample_size=100,
           min_leaf_size=None,
           prediction_type="train",
           features_to_predict=None,
           validation=10,
           importance_tbl=False,
           exp_var_matching=None,
           output_name=None,
           gis=None,
           context=None,
           future=False,
           return_tuple=False):
    """
    .. image:: _static/images/forest/forest.png 

    The 'forest' method is a forest-based classification and regression
    task that creates models and generates predictions using an adaptation of
    Leo Breiman's random forest algorithm, which is a supervised machine
    learning method. Predictions can be performed for both categorical
    variables (classification) and continuous variables (regression).
    Explanatory variables can take the form of fields in the attribute
    table of the training features. In addition to validation of model
    performance based on the training data, predictions can be made to
    another feature dataset.

    The following are examples:

        + Given data on occurrence of seagrass, as well as a number of environmental explanatory
          variables represented as both attributes which has been enriched using a multi-variable grid
          to calculate distances to factories upstream and major ports, future seagrass occurrence can
          be predicted based on future projections for those same environmental explanatory variables.
        + Suppose you have crop yield data at hundreds of farms across the country along with other
          attributes at each of those farms (number of employees, acreage, and so on). Using these
          pieces of data, you can provide a set of features representing farms where you don't have
          crop yield (but you do have all of the other variables), and make a prediction about crop
          yield.
        + Housing values can be predicted based on the prices of houses that have been sold in the
          current year. The sale price of homes sold along with information about the number of
          bedrooms, distance to schools, proximity to major highways, average income, and crime counts
          can be used to predict sale prices of similar homes.

    .. note::
        Forest Based Classification and Regression is available at ArcGIS Enterprise 10.7.

    =========================================================================   ===========================================================================
    **Argument**                                                                **Description**
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    input_layer                                                                 Required layer. The features that will be used to train the dataset. 
                                                                                This layer must include fields representing the variable to predict 
                                                                                and the explanatory variables. See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    var_prediction                                                              Required dict. The variable from the ``input_layer`` parameter
                                                                                containing the values to be used to train the model, and a
                                                                                boolean denoting if it's categorical. This field contains known
                                                                                (training) values of the variable that will be used to predict
                                                                                at unknown locations.

                                                                                Syntax: {"fieldName":"<field name>", "categorical":bool}
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    var_explanatory                                                             Required list. A list of fields representing the explanatory
                                                                                variables and a boolean value denoting whether the fields are
                                                                                categorical. The explanatory variables help predict the value
                                                                                or category of the ``var_prediction`` parameter. Use the
                                                                                categorical parameter for any variables that represent classes
                                                                                or categories (such as land cover or presence or absence).
                                                                                Specify the variable as 'True' for any that represent classes or
                                                                                categories such as land cover or presence or absence and 'False'
                                                                                if the variable is continuous.

                                                                                Syntax: [{"fieldName":"<field name>", "categorical":bool},...]
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    trees                                                                       Required integer. The number of trees to create in the forest model.
                                                                                More trees will generally result in more accurate model
                                                                                prediction, but the model will take longer to calculate.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    max_tree_depth                                                              Optional integer. The maximum number of splits that will be made
                                                                                down a tree. Using a large maximum depth, more splits will be
                                                                                created, which may increase the chances of overfitting the
                                                                                model. The default is data driven and depends on the number of
                                                                                trees created and the number of variables included. 
                                                                                The ``max_tree_depth`` must be positive and less than or equal to 30.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    random_vars                                                                 Optional integer. Specifies the number of explanatory variables
                                                                                used to create each decision tree.Each of the decision trees in
                                                                                the forest is created using a random subset of the explanatory
                                                                                variables specified. Increasing the number of variables used in
                                                                                each decision tree will increase the chances of overfitting
                                                                                your model particularly if there is one or a couple dominant
                                                                                variables. A common practice is to use the square root of the
                                                                                total number of explanatory variables (fields, distances, and
                                                                                rasters combined) if your variablePredict is numeric or divide
                                                                                the total number of explanatory variables (fields, distances,
                                                                                and rasters combined) by 3 if ``var_prediction`` is categorical.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    sample_size                                                                 Optional integer. Specifies the percentage of the ``input_layer`` used
                                                                                for each decision tree.
                                                                                Samples for each tree are taken randomly from two-thirds of the
                                                                                data specified.

                                                                                The default is 100 percent of the data.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    min_leaf_size                                                               Optional integer. The minimum number of observations required to
                                                                                keep a leaf (that is the terminal node on a tree without
                                                                                further splits). For very large data,
                                                                                increasing these numbers will decrease the run time of the
                                                                                tool.

                                                                                The default minimum for regression is 5 and
                                                                                the default for classification is 1.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    prediction_type                                                             Optional string. Specifies the operation mode of the tool. The tool can be run to
                                                                                train a model to only assess performance, or train a model and
                                                                                predict features. Prediction types are as follows:
                
                                                                                    + ``Train`` - A model will be trained, but
                                                                                      no predictions will be generated. Use this option to
                                                                                      assess the accuracy of your model before generating
                                                                                      predictions. This option will output model diagnostics in
                                                                                      the messages window and a chart of variable importance.
                                                                                    + ``TrainAndPredict`` - Predictions or classifications will be
                                                                                      generated for features. Explanatory variables must be
                                                                                      provided for both the training features and the features
                                                                                      to be predicted. The output of this option will be a
                                                                                      feature service, model diagnostics, and an optional
                                                                                      table of variable importance.
                
                                                                                The default value is 'Train'.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    features_to_predict (Required if using ``TrainAndPredict``)                 Optional layer. A feature layer representing locations where predictions will be made. 
                                                                                This layer must include explanatory variable fields that correspond to fields used in ``input_layer``. 
                                                                                This parameter is only used when the ``prediction_type`` is ``TrainAndPredict`` and is required in that case. 
                                                                                See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    validation                                                                  Optional integer. Specifies the percentage (between 10 percent
                                                                                and 50 percent) of inFeatures to reserve as the test dataset
                                                                                for validation. The model will be trained without this random
                                                                                subset of data, and the observed values for those features will
                                                                                be compared to the predicted value. 
                                                                                
                                                                                The default value is 10 percent.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    importance_tbl                                                              Optional boolean. Specifies whether an output table will be
                                                                                generated that contains information describing the importance
                                                                                of each explanatory variable used in the model created.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    exp_var_matching                                                            Optional list of dicts. A list of fields representing the explanatory variables and a
                                                                                boolean values denoting if the fields are categorical. The
                                                                                explanatory variables help predict the value or category of the
                                                                                variable_predict. Use the categorical parameter for any
                                                                                variables that represent classes or categories (such as
                                                                                landcover or presence or absence). Specify the variable as
                                                                                'True' for any that represent classes or categories such as
                                                                                landcover or presence or absence and 'False' if the variable is
                                                                                continuous.
                
                                                                                Syntax: [{"fieldName":"<explanatory field name>", "categorical":bool},
                
                                                                                    + fieldname is the name of the field in the ``input_layer`` used
                                                                                      to predict the ``var_prediction``.
                                                                                    + categorical is one of: 'True' or 'False'. A string field should
                                                                                      always be 'True', and a continue value should always be set as 'False'.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    output_name                                                                 Optional string. The task will create a feature service of the
                                                                                results. You define the name of the service.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    gis                                                                         Optional GIS. The GIS on which this tool runs. If not
                                                                                specified, the active GIS is used.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    context                                                                     Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are four settings:
 
                                                                                #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                                                #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                                                                                #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                                                                #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.  
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    future                                                                      Optional boolean. If 'True', a GPJob is returned instead of
                                                                                results. The GPJob can be queried on the status of the execution.

                                                                                The default value is 'False'.
    -------------------------------------------------------------------------   ---------------------------------------------------------------------------
    return_tuple                                                                Optional boolean. If 'True', a named tuple with multiple output keys is returned.
                                 
                                                                                The default value is 'False'. 
    =========================================================================   ===========================================================================

    :returns: a named tuple with the following keys if ``return_tuple`` is set to 'True':

        "output_trained" : featureLayer

        "output_predicted" : featureLayer	

        "variable_of_importance" : Table 

        "process_info" : list

    else returns a feature layer of the results.

    .. code-block:: python

            # Usage Example: To predict the number of 911 calls in each block group.
            predicted_result = forest(input_layer=call_lyr,
                                      var_prediction={"fieldName":"Calls", "categorical":False},
                                      var_explanatory=[{"fieldName":"Pop", "categorical":False},
                                                        {"fieldName":"Unemployed", "categorical":False},
                                                        {"fieldName":"AlcoholX", "categorical":False},
                                                        {"fieldName":"UnEmpRate", "categorical":False},
                                                        {"fieldName":"MedAge00", "categorical":False}],
                                      trees=50,
                                      max_tree_depth=10,
                                      random_vars=3,
                                      sample_size=100,
                                      min_leaf_size=5,
                                      prediction_type='TrainAndPredict',
                                      validation=10,
                                      importance_tbl=True,
                                      output_name='train and predict number of 911 calls')

    """
    allowed_prediction_types = {
        'train' : "Train",
        'trainandpredict' : 'TrainAndPredict'

    }
    if str(prediction_type).lower() not in allowed_prediction_types:
        raise ValueError("Invalid Prediction type.")
    else:
        prediction_type = allowed_prediction_types[prediction_type.lower()]

    kwargs=locals()

    gis=_arcgis.env.active_gis if gis is None else gis

    if gis.version < [7]:
        return None
    url=gis.properties.helperServices.geoanalytics.url

    params={}
    for key, value in kwargs.items():
        if value is not None:
            params[key]=value

    if output_name is None:
        output_service_name='Forest Based Regression_' + _id_generator()
        output_name=output_service_name.replace(' ', '_')
    else:
        output_service_name=output_name.replace(' ', '_')

    output_service=_create_output_service(gis, output_name, output_service_name, 'Forest Based Classification And Regression')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)



    param_db={
        "input_layer": (_FeatureSet, "inFeatures"),
        "prediction_type" : (str, "predictionType"),
        "features_to_predict" : (_FeatureSet, "featuresToPredict"),
        "var_prediction" : (dict, "variablePredict"),
        "var_explanatory" : (list, "explanatoryVariables"),
        "exp_var_matching" : (list, "explanatoryVariableMatching"),
        "trees" : (int, "numberOfTrees"),
        "max_tree_depth" : (int, "maximumTreeDepth"),
        "min_leaf_size" : (int, "minimumLeafSize"),
        "sample_size" : (int, "sampleSize"),
        "random_vars" : (int, "randomVariables"),
        "validation" : (float, "percentageForValidation"),
        "output_name" : (str, "outputTrainedName"),
        "context": (str, "context"),
        "return_tuple": (bool, "returnTuple"),         
        "importance_tbl" : (bool, "createVariableImportanceTable"),
        "output_trained": (_FeatureSet, "outputTrained"),
        "output_predicted": (_FeatureSet, "outputPredicted"),
        "variable_of_importance": (_Table, "variableOfImportance"),
        "process_info": (list, "processInfo")
    }
    return_values=[
        {"name": 'output_trained', "display_name": "Output Features", "type": _FeatureSet},
        {"name" : "output_predicted", "display_name" : "Output Predicted", "type" : _FeatureSet},
        {"name" : "variable_of_importance", "display_name" : "Variable of Importance", "type" : _Table},
        {"name" : "process_info", "display_name" : "Process Information", "type" : list}
    ]
    if features_to_predict is None and prediction_type == 'TrainAndPredict':
        params["features_to_predict"] = input_layer
        #param_db.pop("features_to_predict")
    try:
        if future:
            gpjob = _execute_gp_tool(gis, "ForestBasedClassificationAndRegression", params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        res = _execute_gp_tool(gis, "ForestBasedClassificationAndRegression", params, param_db, return_values, _use_async, url, True, future=future)

        if return_tuple:
            return res
        else:
            return output_service   
    except:
        output_service.delete()
        raise

    return
#--------------------------------------------------------------------------
def glr(input_layer,
        var_dependent,
        var_explanatory,
        regression_family="Continuous",
        features_to_predict=None,
        gen_coeff_table=False,
        exp_var_matching=None,
        dep_mapping=None,
        output_name=None,
        gis=None,
        context=None,
        future=False,
        return_tuple=False):
    """
    .. image:: _static/images/glr/glr.png 

    This tool performs Generalized Linear Regression (``glr``) to generate
    predictions or to model a dependent variable's relationship to a set of
    explanatory variables. This tool can be used to fit continuous
    (Gaussian/OLS), binary (logistic), and count (Poisson) models.

    The following are examples of the tool's utility:

        + What demographic characteristics contribute to high rates of public transportation usage?
        + Is there a positive relationship between vandalism and burglary?
        + Which variables effectively predict 911 call volume? Given future projections, what is the expected demand for emergency response resources?
        + What variables affect low birth rates?

    ==========================   ===============================================================
    **Argument**                 **Description**
    --------------------------   ---------------------------------------------------------------
    input_layer                  Required layer. The layer containing the dependent and
                                 independent variables. See :ref:`Feature Input<FeatureInput>`.
    --------------------------   ---------------------------------------------------------------
    var_dependent                Required string. The numeric field containing the observed
                                 values you want to model.
    --------------------------   ---------------------------------------------------------------
    var_explanatory              Required list of strings. One or more fields representing independent
                                 explanatory variables in your regression model.
    --------------------------   ---------------------------------------------------------------
    regression_family            Optional string. This field specifies the type of data you are
                                 modeling.

                                 regression_family is one of the following:

                                    + ``Continuous`` - The dependent_variable is continuous. The
                                      model used is Gaussian, and the tool performs
                                      ordinary least squares regression.
                                    + ``Binary`` - The dependent_variable represents presence or
                                      absence. Values must be 0 (absence) or 1 (presence)
                                      values, or mapped to 0 and 1 values using the
                                      parameter.
                                    + ``Count`` - The dependent_variable is discrete and represents
                                      events, such as crime counts, disease incidents,
                                      or traffic accidents. The model used is Poisson
                                      regression.

                                 The default value is 'Continuous'.             
    --------------------------   ---------------------------------------------------------------
    features_to_predict          Optional layer. A layer containing features representing
                                 locations where estimates should be computed. Each feature in
                                 this dataset should contain values for all the explanatory
                                 variables specified. The dependent variable for these features
                                 will be estimated using the model calibrated for the input
                                 layer data. See :ref:`Feature Input<FeatureInput>`.
    --------------------------   ---------------------------------------------------------------
    gen_coeff_table              Optional boolean. Determines if a table with coefficient values
                                 will be returned. By default, the coefficient table is not
                                 returned.
    --------------------------   ---------------------------------------------------------------
    exp_var_matching             Optional list of dicts. A list of the ``var_explanatory`` specified from
                                 the ``input_layer`` and their corresponding fields from the
                                 ``features_to_predict``. By default, if an ``var_explanatory`` variiables is
                                 not mapped, it will match to a field with the same name in the
                                 ``features_to_predict``. This parameter is only used if there is a
                                 ``features_to_predict`` input. You do not need to use it if the
                                 names and types of the fields match between your two input
                                 datasets.

                                 Syntax: [{"predictionLayerField":"<field name>","trainingLayerField": "<field name>"},...]

                                    + predictionLayerField is the name of a field specified in the
                                      var_explanatoryiables parameter.
                                    + trainingLayerField is the field that will match to the field
                                      in the var_explanatoryiables parameter.
    --------------------------   ---------------------------------------------------------------
    dep_mapping                  Optional list of dicts. A list representing the values used to map to 0
                                 (absence) and 1 (presence) for binary regression.

                                 Syntax: [{"value0":"<false value>"},{"value1":"<true value>"}]

                                    + value0 is the string that will be used to represent 0
                                      (absence values).
                                    + value1 is the string that will be used to represent 1
                                      (presence values).

    --------------------------   ---------------------------------------------------------------
    output_name                  Optional string. The task will create a feature service of the
                                 results. You define the name of the service.
    --------------------------   ---------------------------------------------------------------
    gis                          Optional GIS. The GIS on which this tool runs. If not
                                 specified, the active GIS is used.
    --------------------------   ---------------------------------------------------------------
    context                      Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                                 #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                 #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                                 #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                 #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    --------------------------   ---------------------------------------------------------------
    future                       Optional boolean. If 'True', a GPJob is returned instead of
                                 results. The GPJob can be queried on the status of the execution.

                                 The default value is 'False'. 
    --------------------------   ---------------------------------------------------------------      
    return_tuple                 Optional boolean. If 'True', a named tuple with multiple output keys is returned.
                                 
                                 The default value is 'False'. 
    ==========================   ===============================================================

    :returns: a named tuple with the following keys if ``return_tuple`` is set to 'True':

      "output" : featureLayer

      "output_predicted" : featureLayer	

      "coefficient_table" : Table 

      "process_info" : list

    else returns a feature layer of the results.
    .. code-block:: python

            # Usage Example: To train a model for predicting 911 calls.
            
            result_predicted = glr(input_layer=911_calls_lyr,
                                   var_dependent='Calls',
                                   var_explanatory='Unemployed, AlcoholX, UnEmpRate, MedAge00',
                                   regression_family='Count',
                                   gen_coeff_table=True,
                                   output_name="predicted calls")

    """


    _allowed_regression_family = {
        "continuous" : "Continuous",
        "binary" : "Binary",
        "count" : "Count"
    }
    kwargs=locals()

    if regression_family.lower() in _allowed_regression_family:
        regression_family = _allowed_regression_family[regression_family.lower()]
        if 'regression_family' in kwargs:
            kwargs['regression_family'] = _allowed_regression_family[regression_family.lower()]
    else:
        raise ValueError("Invalid regression_family.")

    gis=_arcgis.env.active_gis if gis is None else gis

    if gis.version < [7]:
        return None
    url=gis.properties.helperServices.geoanalytics.url

    params={}
    for key, value in kwargs.items():
        if value is not None:
            params[key]=value

    if isinstance(var_explanatory, list):
        var_explanatory = ', '.join(var_explanatory)
        params["var_explanatory"] = var_explanatory
        

    if output_name is None:
        output_service_name='GLR_' + _id_generator()
        output_name=output_service_name.replace(' ', '_')
    else:
        output_service_name=output_name.replace(' ', '_')

    output_service=_create_output_service(gis, output_name, output_service_name, 'Generalized Linear Regression')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db={
        "input_layer": (_FeatureSet, "inputLayer"),
        "regression_family" : (str, "regressionFamily"),
        "gen_coeff_table" : (bool, "generateCoefficientTable"),
        "exp_var_matching" : (list, "explanatoryVariableMatching"),
        "var_dependent" : (str, "dependentVariable"),
        "var_explanatory" : (list, "explanatoryVariables"),
        "features_to_predict" : (_FeatureSet, "featuresToPredict"),
        "dep_mapping" : (list, "dependentMapping"),
        "output_name" : (str, "outputName"),
        "context": (str, "context"),
        "return_tuple": (bool, "returnTuple"),
        "output": (_FeatureSet, "output"),
        "output_predicted": (_FeatureSet, "outputPredicted"),
        "coefficient_table" : (_Table, "coefficientTable"),
        "process_info" : (list, "processInfo")
    }
    return_values=[
        {"name": 'output', "display_name": "Output Features", "type": _FeatureSet},
        {"name" : "output_predicted", "display_name" : "Output Predicted", "type" : _FeatureSet},
        {"name" : "coefficient_table", "display_name" : "Coefficient Table", "type" : _Table},
        {"name" : "process_info", "display_name" : "Process Information", "type" : list}
    ]

    try:
        if future:
            gpjob = _execute_gp_tool(gis, "GeneralizedLinearRegression", params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        res = _execute_gp_tool(gis, "GeneralizedLinearRegression", params, param_db, return_values, _use_async, url, True, future=future)

        if return_tuple:
            return res
        else:
            return output_service   
    except:
        output_service.delete()
        raise

    return

#--------------------------------------------------------------------------
def find_point_clusters(
    input_layer,
    method,
    min_feature_clusters,
    search_distance=None,
    distance_unit=None,
    output_name=None,
    gis=None,
    context=None,
    future=False,
    time_method=None,
    search_duration=None,
    duration_unit=None):
    """
    This tool extracts clusters from your input point features and identifies any surrounding noise.

    For example, a nongovernmental organization is studying a particular pest-borne disease. It has
    a point dataset representing households in a study area, some of which are infested, and some of
    which are not. By using the Find Point Clusters tool, an analyst can determine clusters of
    infested households to help pinpoint an area to begin treatment and extermination of pests.

    ==========================   ===============================================================
    **Argument**                 **Description**
    --------------------------   ---------------------------------------------------------------
    input_layer                  required FeatureSet, The table, point, line or polygon features
                                 containing potential incidents.
    --------------------------   ---------------------------------------------------------------
    method                       required String. The algorithm used for cluster analysis. This
                                 parameter must be specified as DBSCAN or HDBSCAN.
    --------------------------   ---------------------------------------------------------------
    min_feature_clusters         optional Integer. Minimum number of clusters to find in a dataset.
    --------------------------   ---------------------------------------------------------------
    search_distance              optional Float.  The distance to search between points to form
                                 a cluster.  This is required for DBSCAN.
    --------------------------   ---------------------------------------------------------------
    distance_unit                optional String. The `search_distance` units.
    --------------------------   ---------------------------------------------------------------
    output_name                  optional string, The task will create a feature service of the
                                 results. You define the name of the service.
    --------------------------   ---------------------------------------------------------------
    gis                          optional GIS, the GIS on which this tool runs. If not
                                 specified, the active GIS is used.
    --------------------------   ---------------------------------------------------------------
    context                      Optional dict. The context parameter contains additional settings 
                                 that affect task execution. For this task, there are four settings:

                                 #. Extent (``extent``) - A bounding box that defines the analysis area. 
                                    Only those features that intersect the bounding box will be analyzed.
                                 #. Processing spatial reference (``processSR``) - The features will be 
                                    projected into this coordinate system for analysis.
                                 #. Output spatial reference (``outSR``) - The features will be projected 
                                    into this coordinate system after the analysis to be saved. 
                                    The output spatial reference for the spatiotemporal big data store is always WGS84.
                                 #. Data store (``dataStore``) - Results will be saved to the 
                                    specified data store. The default is the spatiotemporal big data store.
    --------------------------   ---------------------------------------------------------------    
    future                       Optional boolean. If True, a GPJob is returned instead of
                                 results. The GPJob can be queried on the status of the execution.
    --------------------------   ---------------------------------------------------------------    
    time_method                  Optional String. When this parameter is set to Linear and `method` 
                                 is `DBSCAN`, both space and time will be used to find point clusters. 
                                 If `method` is `HDBSCAN`, this parameter will be ignored and clusters 
                                 will be found in space only. This parameter can only be used if 
                                 `input_layer` has time enabled and is of type instant. Temporal 
                                 clustering is available at ArcGIS Enterprise 10.8.
    --------------------------   ---------------------------------------------------------------    
    search_duration              Optional String. When using DBSCAN with timeMethod set as Linear, 
                                 this parameter is the time duration within which 
                                 `min_feature_clusters` must be found. This parameter is not used
                                 when HDBSCAN is chosen as the clustering method or when 
                                 `time_method` is not used.
    --------------------------   ---------------------------------------------------------------    
    duration_unit                Optional String. The units used for the `search_duration` 
                                 parameter. This parameter is required when using DBSCAN but will
                                 not be used with HDBSCAN or space-only DBSCAN.
    ==========================   ===============================================================

    :returns:
       Output feature layer item

    """
    kwargs=locals()

    gis=_arcgis.env.active_gis if gis is None else gis
    url=gis.properties.helperServices.geoanalytics.url

    params={}
    for key, value in kwargs.items():
        if value is not None:
            params[key]=value

    if output_name is None:
        output_service_name='Find Point Clusters_' + _id_generator()
        output_name=output_service_name.replace(' ', '_')
    else:
        output_service_name=output_name.replace(' ', '_')

    output_service=_create_output_service(gis, output_name, output_service_name, 'Find Point Clusters')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db={
        "input_layer": (_FeatureSet, "inputLayer"),
        "method" : (str, "clusterMethod"),
        "min_feature_clusters": (int, "minFeaturesCluster"),
        "distance_unit": (str, "searchDistanceUnit"),
        "search_distance" : (float, "searchDistance"),
        "output_name": (str, "outputName"),
        "time_method" : (str, "timeMethod"),
        "search_duration" : (str, "searchDuration"),
        "duration_unit" : (str, "searchDurationUnit"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values=[
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        if future:
            gpjob = _execute_gp_tool(gis, "FindPointClusters", params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, "FindPointClusters", params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise

    return
#--------------------------------------------------------------------------
def calculate_density(
    input_layer,
    fields=None,
    weight="""Uniform""",
    bin_type="""Square""",
    bin_size=None,
    bin_size_unit=None,
    time_step_interval=None,
    time_step_interval_unit=None,
    time_step_repeat_interval=None,
    time_step_repeat_interval_unit=None,
    time_step_reference=None,
    radius=None,
    radius_unit=None,
    area_units="""SquareKilometers""",
    output_name=None,
    gis=None,
    context=None,
    future=False):
    """
    .. image:: _static/images/calculate_density/calculate_density.png

    The ``calculate_density`` tool creates a density map from point features by spreading known quantities of some
    phenomenon (represented as attributes of the points) across the map. The result is a layer of areas classified
    from least dense to most dense.

    For point input, each point should represent the location of some event or incident, and the result layer
    represents a count of the incident per unit area. A higher density value in a new location means that there
    are more points near that location. In many cases, the result layer can be interpreted as a risk surface
    for future events. For example, if the input points represent locations of lightning strikes, the result
    layer can be interpreted as a risk surface for future lightning strikes.

    Other use cases of this tool include the following:

    * Creating crime density maps to help police departments properly allocate resources to high crime areas.
    * Calculating densities of hospitals within a county. The result layer will show areas with high and low accessibility to
      ospitals, and this information can be used to decide where new hospitals should be built.
    * Identifying areas that are at high risk of forest fires based on historical locations of forest fires.
    * Locating communities that are far from major highways in order to plan where new roads should be constructed.

    =================================================     ========================================================================
    **Argument**                                          **Description**
    -------------------------------------------------     ------------------------------------------------------------------------
    input_layer                                           Required point feature layer. The point layer on which the density will be calculated.

                                                          Analysis using ``Square`` or ``Hexagon`` bins requires a projected coordinate system.
                                                          When aggregating layers into bins, the input layer or processing extent (``processSR``) must
                                                          have a projected coordinate system. At 10.5.1, 10.6, and 10.6.1, if a projected coordinate
                                                          system is not specified when running analysis, the World Cylindrical Equal
                                                          Area (WKID 54034) projection will be used. At 10.7 or later, if a projected coordinate system
                                                          is not specified when running analysis, a projection will be picked based on the extent of the data.
                                                          See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------------------     ------------------------------------------------------------------------
    fields                                                Optional string. Provides one or more field specifying the number of incidents at each location.
                                                          You can calculate the density on multiple fields, and the count of points will always have the density calculated.
    -------------------------------------------------     ------------------------------------------------------------------------
    weight                                                Required string. The type of weighting applied to the density calculation. There are two options:

                                                            * ``Uniform`` - Calculates a magnitude-per-area. This is the default.
                                                            * ``Kernel`` - Applies a kernel function to fit a smooth tapered surface to each point.

                                                          The default value is "Uniform".
    -------------------------------------------------     ------------------------------------------------------------------------
    bin_type                                              Required string. The type of bin used to calculate density.

                                                          Choice list: ['Hexagon', 'Square'].
    -------------------------------------------------     ------------------------------------------------------------------------
    bin_size                                              Required float. The distance for the bins that the ``input_layer`` will be analyzed using.
                                                          When generating bins, for Square, the number and units specified determine the
                                                          height and length of the square. For ``Hexagon``, the number and units specified
                                                          determine the distance between parallel sides.
    -------------------------------------------------     ------------------------------------------------------------------------
    bin_size_unit                                         Required string. The distance unit for the bins for which the density will be calculated.
                                                          The linear unit to be used with the value specified in ``bin_size``.

                                                          The default is 'Meters'.
    -------------------------------------------------     ------------------------------------------------------------------------
    time_step_interval                                    Optional integer. A numeric value that specifies duration of the time step interval. This option is
                                                          only available if the input points are time-enabled and represent an instant in time.

                                                          The default value is 'None'.
    -------------------------------------------------     ------------------------------------------------------------------------
    time_step_interval_unit                               Optional string. A string that specifies units of the time step interval.
                                                          This option is only available if the input points are time-enabled and represent an instant in time.

                                                          Choice list: ['Milliseconds', 'Seconds', 'Minutes', 'Hours', 'Days', 'Weeks', 'Months', 'Years']

                                                          The default value is 'None'.
    -------------------------------------------------     ------------------------------------------------------------------------
    time_step_repeat_interval                             Optional integer. A numeric value that specifies how often the time step repeat occurs.
                                                          This option is only available if the input points are time-enabled and of time type instant.
    -------------------------------------------------     ------------------------------------------------------------------------
    time_step_repeat_interval_unit                        Optional string. A string that specifies the temporal unit of the step repeat.
                                                          This option is only available if the input points are time-enabled and of time type instant.

                                                          Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']

                                                          The default value is 'None'.
    -------------------------------------------------     ------------------------------------------------------------------------
    time_step_reference                                   Optional datetime. A date that specifies the reference time to align the time slices to,
                                                          represented in milliseconds from epoch. If time_step_reference is set
                                                          to 'None', time stepping will align to January 1st, 1970 (datetime(1970, 1, 1)).
                                                          This option is only available if the input points are time-enabled and of time type instant.

                                                          The default value is 'None'.
    -------------------------------------------------     ------------------------------------------------------------------------
    radius                                                Required integer. The size of the neighborhood within which to calculate the density.
                                                          The radius size must be larger than the ``bin_size``.
    -------------------------------------------------     ------------------------------------------------------------------------
    radius_unit                                           Required string. The distance unit for the radius defining the neighborhood for which the density will be calculated.
                                                          The linear unit to be used with the value specified in ``bin_size``.

                                                          Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']

                                                          The default value is 'Meters'.
    -------------------------------------------------     ------------------------------------------------------------------------
    area_units                                            Optional string. The desired output units of the density values. If density values are very small, you can increase the
                                                          size of the area units (for example, square meters to square kilometers) to return larger values.
                                                          This value only scales the result. Possible area units are:

                                                          Choice list: ['SquareMeters', 'SquareKilometers', 'Hectares', 'SquareFeet', 'SquareYards', 'SquareMiles', 'Acres'].

                                                          The default value is "SquareKilometers".
    -------------------------------------------------     ------------------------------------------------------------------------
    output_name                                           Optional string. The method will create a feature service of the results. You define the name of the service.
    -------------------------------------------------     ------------------------------------------------------------------------
    gis                                                   Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------------------------------     ------------------------------------------------------------------------
    context                                               Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                                                          #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                          #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                                                          #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                                          #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    -------------------------------------------------     ------------------------------------------------------------------------
    future                                                Optional boolean. If True, a GPJob is returned instead of
                                                          results. The GPJob can be queried on the status of the execution.
    =================================================     ========================================================================

    :returns: result_layer : Output Features as feature layer item.

    .. code-block:: python

            # Usage Example: Aggregate the number of Hurricanes within 1 meter to calculate density of Hurricane damage.

            cal_den_result = calculate_density(input_layer=hurricane_lyr,
                                               fields='Damage',
                                               weight='Uniform',
                                               bin_type='Square',
                                               bin_size=1,
                                               bin_size_unit="Meters",
                                               radius=2,
                                               radius_unit="Yards")

    """
    kwargs=locals()

    gis=_arcgis.env.active_gis if gis is None else gis
    url=gis.properties.helperServices.geoanalytics.url

    params={}
    for key, value in kwargs.items():
        if value is not None:
            params[key]=value

    if output_name is None:
        output_service_name='Calculate Density Analysis_' + _id_generator()
        output_name=output_service_name.replace(' ', '_')
    else:
        output_service_name=output_name.replace(' ', '_')

    output_service=_create_output_service(gis, output_name, output_service_name, 'Calculate Density')

    params['output_name']=_json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db={
        "input_layer": (_FeatureSet, "inputLayer"),
        "fields": (str, "fields"),
        "weight": (str, "weight"),
        "bin_type": (str, "binType"),
        "bin_size": (float, "binSize"),
        "bin_size_unit": (str, "binSizeUnit"),
        "time_step_interval": (int, "timeStepInterval"),
        "time_step_interval_unit": (str, "timeStepIntervalUnit"),
        "time_step_repeat_interval": (int, "timeStepRepeatInterval"),
        "time_step_repeat_interval_unit": (str, "timeStepRepeatIntervalUnit"),
        "time_step_reference": (_datetime, "timeStepReference"),
        "radius": (float, "radius"),
        "radius_unit": (str, "radiusUnit"),
        "area_units": (str, "areaUnits"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values=[
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        if future:
            gpjob = _execute_gp_tool(gis, "CalculateDensity", params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, "CalculateDensity", params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise


calculate_density.__annotations__={
    'fields': str,
    'weight': str,
    'bin_type': str,
    'bin_size': float,
    'bin_size_unit': str,
    'time_step_interval': int,
    'time_step_interval_unit': str,
    'time_step_repeat_interval': int,
    'time_step_repeat_interval_unit': str,
    'time_step_reference': _datetime,
    'radius': float,
    'radius_unit': str,
    'area_units': str,
    'output_name': str}

#--------------------------------------------------------------------------
def find_hot_spots(
    point_layer,
    bin_size=5,
    bin_size_unit="Miles",
    neighborhood_distance=5,
    neighborhood_distance_unit="Miles",
    time_step_interval=None,
    time_step_interval_unit=None,
    time_step_alignment=None,
    time_step_reference=None,
    output_name=None,
    gis=None,
    context=None,
    future=False):
    """
    .. image:: _static/images/geo_find_hot_spots/geo_find_hot_spots.png

    The ``find_hot_spots`` tool analyzes point data (such as crime incidents, traffic accidents, trees, and so on)
    or field values associated with points. It finds statistically significant spatial clusters of high incidents (hot spots)
    and low incidents (cold spots). Hot spots are locations with lots of points and cold spots are locations with very few points.

    The result map layer shows hot spots in red and cold spots in blue. The darkest red features indicate the strongest
    clustering of point densities; you can be 99 percent confident that the clustering associated with these features
    could not be the result of random chance. Similarly, the darkest blue features are associated with the strongest
    spatial clustering of the lowest point densities. Features that are beige are not part of a statistically
    significant cluster; the spatial pattern associated with these features could very likely be the result of random
    processes and random chance.

    ==============================================================================================  ===============================================================
    **Argument**                                                                                    **Description**
    ----------------------------------------------------------------------------------------------  ---------------------------------------------------------------
    point_layer                                                                                     Required feature layer. The point feature layer for which hot spots will be calculated.
                                                                                                    See :ref:`Feature Input<FeatureInput>`.

                                                                                                    .. Note::
                                                                                                        Analysis using bins requires a projected coordinate system. When aggregating layers into bins,
                                                                                                        the input layer or processing extent (``processSR``) must have a projected coordinate system.
                                                                                                        At 10.5.1, 10.6, and 10.6.1, if a projected coordinate system is not specified when running
                                                                                                        analysis, the World Cylindrical Equal Area (WKID 54034) projection will be used. At 10.7 or later,
                                                                                                        if a projected coordinate system is not specified when running analysis, a projection will be
                                                                                                        picked based on the extent of the data.
    ----------------------------------------------------------------------------------------------  ---------------------------------------------------------------
    bin_size                                                                                        Optional float. The distance for the square bins the ``point_layer`` will be aggregated into.
    ----------------------------------------------------------------------------------------------  ---------------------------------------------------------------
    bin_size_unit                                                                                   Optional string. The distance unit for the bins with which hot spots will be calculated.
                                                                                                    The linear unit to be used with the value specified in ``bin_size``.
                                                                                                    When generating bins the number and units specified determine the height and length of the square.

                                                                                                    Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']

                                                                                                    The default value is ``Miles``.
    ----------------------------------------------------------------------------------------------  ---------------------------------------------------------------
    neighborhood_distance                                                                           Optional float. The size of the neighborhood within which to calculate the hot spots.
                                                                                                    The radius size must be larger than ``bin_size``.
    ----------------------------------------------------------------------------------------------  ---------------------------------------------------------------
    neighborhood_distance_unit                                                                      Optional string. The distance unit for the radius defining the neighborhood where the hot spots
                                                                                                    will be calculated. The linear unit to be used with the value specified in ``bin_size``.

                                                                                                    Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']

                                                                                                    The default value is 'Miles'.
    ----------------------------------------------------------------------------------------------  ---------------------------------------------------------------
    time_step_interval                                                                              Optional integer. A numeric value that specifies duration of the time step interval.
                                                                                                    This option is only available if the input points are time-enabled and represent an instant in time.
    ----------------------------------------------------------------------------------------------  ---------------------------------------------------------------
    time_step_interval_unit                                                                         Optional string. A string that specifies units of the time step interval.
                                                                                                    This option is only available if the input points are time-enabled and represent an instant in time.

                                                                                                    Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']
    ----------------------------------------------------------------------------------------------  ---------------------------------------------------------------
    time_step_alignment                                                                             Optional string. Defines how aggregation will occur based on a given ``time_step_interval``. Options are as follows:

                                                                                                    Choice list:['EndTime', 'StartTime', 'ReferenceTime']

                                                                                                        * ``StartTime`` - Time is aligned to the first feature in time.
                                                                                                        * ``EndTime`` - Time is aligned to the last feature in time.
                                                                                                        * ``ReferenceTime`` - Time is aligned a specified time in ``time_step_reference``.
    ----------------------------------------------------------------------------------------------  ---------------------------------------------------------------
    time_step_reference (Required if ``time_step_alignment`` is ReferenceTime)                      Optional datetime. A date that specifies the reference time to align the time slices to.
                                                                                                    This option is only available if the input points are time-enabled and of time type instant.
    ----------------------------------------------------------------------------------------------  ---------------------------------------------------------------
    output_name                                                                                     Optional string. The task will create a feature service of the results. You define the name of the service.
    ----------------------------------------------------------------------------------------------  ---------------------------------------------------------------
    context                                                                                         Optional string. Context contains additional settings that affect task execution. For this task, there are three settings:

                                                                                                    #.  Extent (``extent``) - a bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                                                                    #. Processing spatial reference (``processSR``) The features will be projected into this coordinate system for analysis.
                                                                                                    #. Output Spatial Reference (``outSR``) - the features will be projected into this coordinate system after the analysis to be saved.
                                                                                                       The output spatial reference for the spatiotemporal big data store is always WGS84.
                                                                                                    #. Data store (``dataStore``) Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    ----------------------------------------------------------------------------------------------  ---------------------------------------------------------------
    gis                                                                                             Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    ----------------------------------------------------------------------------------------------  ---------------------------------------------------------------
    future                                                                                          Optional boolean. If True, a GPJob is returned instead of results. The GPJob can be queried on the status of the execution.
    ==============================================================================================  ===============================================================

    :returns: Output Features as a feature layer collection item

    .. code-block:: python

        # Usage Example: To find significantly hot or cold spots of fire incidents.
        find_hot_spots(point_layer=fire,
                       bin_size=5,
                       bin_size_unit='Miles',
                       neighborhood_distance=5,
                       neighborhood_distance_unit='Miles',
                       time_step_interval=1,
                       time_step_interval_unit="Years",
                       time_step_alignment="StartTime",
                       time_step_reference=None,
                       output_name='find hot spots')


    """
    kwargs=locals()

    gis=_arcgis.env.active_gis if gis is None else gis
    url=gis.properties.helperServices.geoanalytics.url

    params={}
    for key, value in kwargs.items():
        if value is not None:
            params[key]=value

    if output_name is None:
        output_service_name='Hotspot Analysis_' + _id_generator()
        output_name=output_service_name.replace(' ', '_')
    else:
        output_service_name=output_name.replace(' ', '_')

    output_service=_create_output_service(gis, output_name, output_service_name, 'Find Hotspots')

    params['output_name']=_json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db={
        "point_layer": (_FeatureSet, "pointLayer"),
        "bin_size": (float, "binSize"),
        "bin_size_unit": (str, "binSizeUnit"),
        "neighborhood_distance": (float, "neighborhoodDistance"),
        "neighborhood_distance_unit": (str, "neighborhoodDistanceUnit"),
        "time_step_interval": (int, "timeStepInterval"),
        "time_step_interval_unit": (str, "timeStepIntervalUnit"),
        "time_step_alignment": (str, "timeStepAlignment"),
        "time_step_reference": (_datetime, "timeStepReference"),
        #"cell_size" : (int, "cellSize"),
        #"cell_size_units": (str, "cellSizeUnits"),
        #"shape_type" : (str, "shapeType"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values=[
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        if future:
            gpjob = _execute_gp_tool(gis, "FindHotSpots", params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, "FindHotSpots", params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise

find_hot_spots.__annotations__={
    'bin_size': float,
    'bin_size_unit': str,
    'neighborhood_distance': float,
    'neighborhood_distance_unit': str,
    'time_step_interval': int,
    'time_step_interval_unit': str,
    'time_step_alignment': str,
    'time_step_reference': _datetime,
    'output_name': str}

#--------------------------------------------------------------------------
def create_space_time_cube(point_layer: _FeatureSet,
                           bin_size: float,
                           bin_size_unit: str,
                           time_step_interval: int,
                           time_step_interval_unit: str,
                           time_step_alignment: str=None,
                           time_step_reference: _datetime=None,
                           summary_fields: str=None,
                           output_name: str=None,
                           context: str=None,
                           gis=None,
                           future: bool=False) -> DataFile:
    """
    .. image:: _static/images/create_space_time_cube/create_space_time_cube.png 
    ``create_space_time_cube`` works with a layer of point features that are time enabled. 
    It aggregates the data into a three-dimensional cube of space-time bins. 
    When determining the point in a space-time bin relationship, statistics about all 
    points in the space-time bins are calculated and assigned to the bins. 
    The most basic statistic is the number of points within the bins, but you can 
    calculate other statistics as well.
    For example, suppose you have point features of crimes in a city, and you want 
    to summarize the number of crimes in both space and time. You can calculate the 
    space-time cube for the dataset, and use the cube to further analyze trends 
    such as emerging hot and cold spots.
    ============================================================================     ===================================================================================================
    **Argument**                                                                     **Description**
    ----------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------
    point_layer                                                                      Required point feature layer. The point features that will be 
                                                                                     aggregated into the bins specified in geographical size by the ``bin_size`` 
                                                                                     and ``bin_size_unit`` parameters and temporal size by 
                                                                                     the ``time_step_interval`` and ``time_step_interval_unit`` parameters. 
                                                                                     See :ref:`Feature Input<FeatureInput>`.
                                                                                     Analysis using bins requires a projected coordinate system. 
                                                                                     When aggregating layers into bins, the input layer or processing 
                                                                                     extent (``processSR``) must have a projected coordinate system. 
                                                                                     At 10.5.1, 10.6, and 10.6.1, if a projected coordinate system is 
                                                                                     not specified when running analysis, the World Cylindrical Equal 
                                                                                     Area (WKID 54034) projection will be used. At 10.7 or later, if a 
                                                                                     projected coordinate system is not specified when running analysis, 
                                                                                     a projection will be picked based on the extent of the data.
                                                                                     .. Note: The ``input_layer`` must have a minimum of 60 features.
    ----------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------
    bin_size                                                                         Required float. The distance for the bins into which ``point_layer`` will be aggregated.
                                                                                     .. Note: A ``create_space_time_cube`` must have at least 10 time slices.
    ----------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------
    bin_size_unit                                                                    Required string. The distance unit for the bins into which ``point_layer`` will be aggregated.
                                                                                     Choice list: ['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']
    ----------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------
    time_step_interval                                                               Required integer. A numeric value that specifies the duration of the time bin.
                                                                                     .. Note: 
    ----------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------
    time_step_interval_unit                                                          Required string. A numeric value that specifies the duration unit of the time bin.
                                                                                     Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']
    ----------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------
    time_step_alignment                                                              Optional string. Defines how aggregation will occur based on a given timeInterval. Options are as follows:
                                                                                     Choice list: ['EndTime', 'StartTime', 'ReferenceTime']
                                                                                        * ``StartTime`` - Time is aligned to the first feature in time
                                                                                        * ``EndTime`` - Time is aligned to the last feature in time
                                                                                        * ``ReferenceTime`` - Time is aligned a specified time
    ----------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------
    time_step_reference (Required if ``time_step_alignment`` is ReferenceTime)       Optional datetime. A date that specifies the reference time to align the 
                                                                                     time bins to if ReferenceTime is specified in ``time_step_alignment``. 
    ----------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------
    summary_fields                                                                   Optional string. A list of field names, statistical summary types, and the 
                                                                                     fill option for empty values that you want to calculate for all points 
                                                                                     within each space-time bin. Note that the count of points within each 
                                                                                     bin is always returned. By default, all statistics are returned.
                                                                                     Example: [{"statisticType": "statistic type", "onStatisticField": "field name", "fillType": "fill type", "onStatisticField": "fieldName2"}]
                                        
                                                                                     fieldName is the name of the fields in the input point layer.
                                                                                     statisticType is one of the following for numeric fields:
                                                                                     * ``Sum`` - Adds the total value of all the points in each polygon.
                                                                                     * ``Mean`` - Calculates the average of all the points in each polygon.
                                                                                     * ``Min`` - Finds the smallest value of all the points in each polygon.
                                                                                     * ``Max`` - Finds the largest value of all the points in each polygon.
                                                                                     * ``Stddev`` - Finds the standard deviation of all the points in each polygon.
                                                                                     statisticType is the following for string fields:
                                                                                     * ``Count`` - Totals the number of strings for all the points in each polygon.
 
                                                                                     fillType is one of the following:
                                                                                     * ``zeros`` - Fills missing values with zeros. This is most appropriate for fields representing counts.
                                                                                     * ``spatialNeighbors`` - Fills missing values by averaging the spatial neighbors. Neighbors are determined by a second degree queens contiguity.
                                                                                     * ``spaceTimeNeighbors`` - Fills missing values by averaging the space-time neighbors. Neighbors are determined by a second degree queens contiguity in both space and time.
                                                                                     * ``temporalTrend`` - Interpolates values using a univariate spline.   
    ----------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------
    output_name                                                                      Required string. The task will create a space time cube (netCDF) of the results. You define the name of the space time cube.
    ----------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------
    context                                                                          Optional string. Context contains additional settings that affect task execution. For this task, there are three settings:
                                                                                     
                                                                                     #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                                                     #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
    ----------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------
    gis                                                                              Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    ----------------------------------------------------------------------------     ---------------------------------------------------------------------------------------------------
    future                                                                           Optional boolean. If True, a GPJob is returned instead of results. The GPJob can be queried on the status of the execution.
    ============================================================================     ===================================================================================================
    :returns: dict with url containing the path to Output Space Time Cube (netCDF) dataFile. When you browse to the output url, your netCDF will automatically download to your local machine.
    
    .. code-block:: python
            # Usage Example: To aggregate Chicago homicides date layer into 3-dimensional cubes of 5 miles bin.
            create_space_time_cube(point_layer=lyr,
                                   bin_size=5,
                                   bin_size_unit="Miles",
                                   time_step_interval=1,
                                   time_step_interval_unit="Days",
                                   time_step_alignment='StartTime',
                                   time_step_reference=datetime(1995, 10, 4),
                                   summary_fields=[{"statisticType": "Mean", "onStatisticField" : "Beat", "fillType" : "temporalTrend" }],
                                   output_name="spacecube")
    """

    kwargs=locals()

    gis=_arcgis.env.active_gis if gis is None else gis
    url=gis.properties.helperServices.geoanalytics.url

    params={}
    for key, value in kwargs.items():
        if value is not None:
            params[key]=value

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db={
        "point_layer": (_FeatureSet, "pointLayer"),
        "bin_size": (float, "binSize"),
        "bin_size_unit": (str, "binSizeUnit"),
        "time_step_interval": (int, "timeStepInterval"),
        "time_step_interval_unit": (str, "timeStepIntervalUnit"),
        "time_step_alignment": (str, "timeStepAlignment"),
        "time_step_reference": (_datetime, "timeStepReference"),
        "summary_fields": (str, "summaryFields"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output_cube": (DataFile, "Output Space Time Cube"),
    }
    return_values=[
        {"name": "output_cube", "display_name": "Output Space Time Cube", "type": DataFile},
    ]
    if future:
        output_service = None
        gpjob = _execute_gp_tool(gis, "CreateSpaceTimeCube", params, param_db, return_values, _use_async, url, True, future=future)
        return GAJob(gpjob=gpjob, return_service=output_service)
    return _execute_gp_tool(gis, "CreateSpaceTimeCube", params, param_db, return_values, _use_async, url, True, future=future)





from ._overlay import overlay
from ._select import select
from ._ops import contains, crosses
from ._ops import disjoint, equals
from ._ops import overlaps, touches
from ._ops import within

__all__ = ['overlay', 'select','contains', 'crosses',
           'disjoint', 'equals','overlaps', 'touches',
           'within']


import os
import json
import string
import random

import arcgis
from arcgis.gis import Layer
from arcgis.features import FeatureCollection
from arcgis.geoprocessing._job import GPJob

def _id_generator(size=6, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(size))


def _feature_input(self, input_layer):
    input_layer_url = ""
    if isinstance(input_layer, arcgis.gis.Item):
        if input_layer.type.lower() == 'feature service':
            input_param =  {'url': input_layer.layers[0].url}
        elif input_layer.type.lower() == 'feature collection':
            fcdict = input_layer.get_data()
            fc = FeatureCollection(fcdict['layers'][0])
            input_param =  fc.layer
        else:
            raise TypeError("item type must be feature service or feature collection")

    elif isinstance(input_layer, arcgis.features.FeatureLayerCollection):
        input_layer_url = input_layer.layers[0].url
        input_param =  {"url": input_layer_url }

    elif isinstance(input_layer, FeatureCollection):
        input_param =  input_layer.properties

    elif isinstance(input_layer, Layer):
        input_layer_url = input_layer.url
        input_param =  {"url": input_layer_url }

    elif isinstance(input_layer, dict):
        input_param =  input_layer

    elif isinstance(input_layer, str):
        input_layer_url = input_layer
        input_param =  {"url": input_layer_url }

    else:
        raise Exception("Invalid format of input layer. url string, feature service Item, feature service instance or dict supported")

    return input_param


def _set_context(params):
    out_sr = arcgis.env.out_spatial_reference
    process_sr = arcgis.env.process_spatial_reference
    out_extent = arcgis.env.analysis_extent
    output_datastore = arcgis.env.output_datastore
    default_aggregation_styles = arcgis.env.default_aggregation_styles

    context = {}
    set_context = False

    if default_aggregation_styles is not None and\
       isinstance(default_aggregation_styles, bool):
        context['defaultAggregationStyles'] = default_aggregation_styles
        set_context = True
    if out_sr is not None:
        context['outSR'] = {'wkid': int(out_sr)}
        set_context = True
    if out_extent is not None:
        context['extent'] = out_extent
        set_context = True
    if process_sr is not None:
        context['processSR'] = {'wkid': int(process_sr)}
        set_context = True
    if output_datastore is not None:
        context['dataStore'] = output_datastore
        set_context = True

    if set_context:
        params["context"] = json.dumps(context)



def _create_output_service(gis, output_name, output_service_name='Analysis feature service', task='GeoAnalytics'):
    ok = gis.content.is_service_name_available(output_name, 'Feature Service')
    if not ok:
        raise RuntimeError("A Feature Service by this name already exists: " + output_name)
    if arcgis.env.output_datastore is not None:
        output_datastore = arcgis.env.output_datastore
    else:
        output_datastore = "spatiotemporal"
    createParameters = {
            "currentVersion": 10.2,
            "serviceDescription": "",
            "hasVersionedData": False,
            "supportsDisconnectedEditing": False,
            "hasStaticData": True,
            "maxRecordCount": 2000,
            "supportedQueryFormats": "JSON",
            "capabilities": "Query",
            "description": "",
            "copyrightText": "",
            "allowGeometryUpdates": False,
            "syncEnabled": False,
            "editorTrackingInfo": {
                "enableEditorTracking": False,
                "enableOwnershipAccessControl": False,
                "allowOthersToUpdate": True,
                "allowOthersToDelete": True
            },
            "xssPreventionInfo": {
                "xssPreventionEnabled": True,
                "xssPreventionRule": "InputOnly",
                "xssInputRule": "rejectInvalid"
            },
            "tables": [],
            "name": output_service_name.replace(' ', '_'),
            "options": {
                "dataSourceType": output_datastore
            }
        }

    output_service = gis.content.create_service(output_name, create_params=createParameters, service_type="featureService")
    description = "Feature Service generated from running the " + task + " tool."
    item_properties = {
            "description" : description,
            "tags" : "Analysis Result, " + task,
            "snippet": output_service_name
            }
    output_service.update(item_properties)
    return output_service



class GAJob(object):
    """
    Represents a Single GeoAnalytics Job.  The `GAJob` class allows for the asynchronous operation
    of any geoprocessing task.  To request a GAJob task, the code must be called with `future=True`
    or else the operation will occur synchronously.  This class is not intended for users to call
    directly.


    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    gpjob             Required GPJob. The geoprocessing job.
    ----------------  ---------------------------------------------------------------
    return_service    Optional Item. The service to return to the user.
    ================  ===============================================================

    """
    _gpjob = None
    _return_service = None
    #----------------------------------------------------------------------
    def __init__(self, gpjob, return_service=None):
        """
        initializer
        """
        assert isinstance(gpjob, GPJob)
        self._gpjob = gpjob
        self._return_service = return_service
    #----------------------------------------------------------------------
    def __str__(self):
        return "<%s GA Job: %s>" % (self.task, self._gpjob._jobid)
    #----------------------------------------------------------------------
    def __repr__(self):
        return "<%s GA Job: %s>" % (self.task, self._gpjob._jobid)
    #----------------------------------------------------------------------
    @property
    def task(self):
        """Returns the task name.
        :returns: string
        """
        return self._gpjob.task
    #----------------------------------------------------------------------
    @property
    def status(self):
        """
        returns the GP status

        :returns: String
        """
        return self._gpjob.status
    #----------------------------------------------------------------------
    def cancel(self):
        """
        Attempt to cancel the call. If the call is currently being executed
        or finished running and cannot be cancelled then the method will
        return False, otherwise the call will be cancelled and the method
        will return True.

        :returns: boolean
        """
        cancel = self._gpjob.cancel()
        if self._return_service:
            self._return_service.delete()
        return cancel

    #----------------------------------------------------------------------
    def cancelled(self):
        """
        Return True if the call was successfully cancelled.

        :returns: boolean
        """
        return self._gpjob.cancelled()
    #----------------------------------------------------------------------
    def running(self):
        """
        Return True if the call is currently being executed and cannot be cancelled.

        :returns: boolean
        """
        return self._gpjob.running()
    #----------------------------------------------------------------------
    def done(self):
        """
        Return True if the call was successfully cancelled or finished running.

        :returns: boolean
        """
        return self._gpjob.done()
    #----------------------------------------------------------------------
    def result(self):
        """
        Return the value returned by the call. If the call hasn't yet completed
        then this method will wait.

        :returns: object
        """
        res = self._gpjob.result()
        if self._return_service:
            return self._return_service
        else:
            return res


import arcgis
from arcgis._impl.common._utils import _lazy_property
from arcgis.apps.tracker import LocationTrackingError
import re
import math


class TrackView:
    """
    A Track View

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    item                   Required :class:`~arcgis.gis.Item`. The item that represents the
                           Track View.
    ==================     ====================================================================

    .. code-block:: python

        # Get a Track View and list mobile users.

        import arcgis
        gis = arcgis.gis.GIS("https://arcgis.com", "<username>", "<password>")
        item = gis.content.get("<item-id>")
        track_view = arcgis.apps.tracker.TrackView(item)
        mobile_users = track_view.mobile_users.list()

    """
    def __init__(self, item):
        if "Location Tracking View" not in item.typeKeywords:
            raise LocationTrackingError("Item is not a Track View")
        self._item = item
        self._gis = item._gis

    def delete(self):
        """
        Deletes the Track View, including the group and view service.
        """
        if self.group:
            self.group.protected = False
            self.group.delete()
        self._item.protect(False)
        self._item.delete()

    @_lazy_property
    def viewers(self):
        """The :class:`~arcgis.apps.tracker.TrackViewerManager` for the Track View"""
        return TrackViewerManager(self)

    @_lazy_property
    def mobile_users(self):
        """The :class:`~arcgis.apps.tracker.MobileUserManager` for the Track View"""
        return MobileUserManager(self)

    @_lazy_property
    def group(self):
        """The group that contains the Track Viewers and Layers"""
        try:
            return self._gis.groups.get(self.item.properties["trackViewGroup"])
        except:
            return None

    @property
    def item(self):
        """The Track View :class:`~arcgis.gis.Item`"""
        return self._item

    @_lazy_property
    def tracks_layer(self):
        """The tracks :class:`~arcgis.features.FeatureLayer`"""
        return self._item.layers[0]

    @_lazy_property
    def last_known_locations_layer(self):
        """The last known locations :class:`~arcgis.features.FeatureLayer`"""
        return self._item.layers[1]


class TrackViewerManager:
    """
    A class that manages the Track Viewers in the Track View.

    It can be accessed from the TrackView as :py:attr:`~arcgis.apps.tracker.TrackView.viewers`

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    track_view             Required :class:`~arcgis.apps.tracker.TrackView`. The Track View to
                           configure Track Viewers for.
    ==================     ====================================================================
    """
    def __init__(self, track_view):
        self._track_view = track_view

    def add(self, viewers):
        """
        Adds the specified usernames as Track Viewers

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        viewers                Required :class:`List` of :class:`String` or :class:`arcgis.gis.User`.
                               The list of usernames/users to add as Track Viewers.
        ==================     ====================================================================
        """
        if isinstance(viewers, (str, arcgis.gis.User)):
            viewers = [viewers]
        max_add_per_call = 25
        for i in range(0, math.ceil(len(viewers)/max_add_per_call)):
            self._track_view.group.add_users(viewers[i * max_add_per_call:(i * max_add_per_call) + max_add_per_call])

    def delete(self, viewers):
        """
        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        viewers                Required :class:`List` of :class:`String` or :class:`arcgis.gis.User`.
                               The list of usernames/users to remove as Track Viewers.
        ==================     ====================================================================
        """
        if isinstance(viewers, (str, arcgis.gis.User)):
            viewers = [viewers]
        max_add_per_call = 25
        for i in range(0, math.ceil(len(viewers)/max_add_per_call)):
            self._track_view.group.remove_users(viewers[i * max_add_per_call:(i * max_add_per_call) + max_add_per_call])

    def list(self):
        """
        Lists all of the Track Viewers

        :return: :class:`List` of :class:`String` representing usernames
        """
        results = self._track_view.group.get_members()
        usernames = results["users"]
        usernames.append(results["owner"])
        return usernames


class MobileUserManager:
    """
    A class that manages the Mobile Users in the Track View.

    It can be accessed from the TrackView as :py:attr:`~arcgis.apps.tracker.TrackView.mobile_users`

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    track_view             Required :class:`~arcgis.apps.tracker.TrackView`. The Track View to
                           configure Mobile Users for.
    ==================     ====================================================================
    """

    _VDQ_RE = """^created_user\s+(?:in|IN)\s+\(\s*((?:'[^']*')(?:\s*,\s*'[^']*')*)\s*\)(?:\s+((?:and|AND).*))?$"""
    _DEFAULT_VDQ = """created_user in ('')"""

    def __init__(self, track_view):
        self._track_view = track_view

    def add(self, users):
        """
        Adds the specified usernames as Mobile Users

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        users                  Required :class:`List` of :class:`String` or :class:`arcgis.gis.User`.
                               The list of usernames/users to add as Mobile Users
        ==================     ====================================================================
        """
        if isinstance(users, str):
            users = {users}
        elif isinstance(users, arcgis.gis.User):
            users = {users.username}
        elif isinstance(users, (set, list)):
            if len(users) > 0 and isinstance(next(iter(users)), arcgis.gis.User):
                users = {u.username for u in users}
            else:
                users = set(users)
        self._update_vdq(users.union(self.list()))

    def delete(self, users):
        """
        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        users                  Required :class:`List` of :class:`String` or :class:`arcgis.gis.User`.
                               The list of usernames/users to remove as Mobile Users.
        ==================     ====================================================================
        """
        if isinstance(users, str):
            users = {users}
        elif isinstance(users, arcgis.gis.User):
            users = {users.username}
        elif isinstance(users, (set, list)):
            if len(users) > 0 and isinstance(next(iter(users)), arcgis.gis.User):
                users = {u.username for u in users}
        self._update_vdq(set(self.list()).difference(users))

    def list(self):
        """
        Lists all of the Track Viewers

        :return: :class:`List` of :class:`String` representing usernames
        """
        pattern = re.compile(self._VDQ_RE)
        match = pattern.match(self.view_definition_query)
        if match is None:
            raise LocationTrackingError("Unable to parse viewDefinitionQuery")
        return [username.strip()[1:-1] for username in match[1].split(',') if username.strip()[1:-1]]

    def _generate_users_where_clause(self, usernames):
        if len(usernames) > 0:
            return "created_user in ({})".format(','.join(["'{}'".format(u) for u in usernames]))
        return self._DEFAULT_VDQ

    def _update_vdq(self, usernames):
        # we do not validate the usernames because, they could represent people who used to be in
        # the org and have historical tracks
        new_users_clause = self._generate_users_where_clause(usernames)
        pattern = re.compile(self._VDQ_RE)
        match = pattern.match(self.view_definition_query)
        custom_section = match.group(2)
        if custom_section is None:
            new_vdq = new_users_clause
        else:
            new_vdq = "{} {}".format(new_users_clause, custom_section)
        self._track_view.tracks_layer.manager.update_definition({
            "viewDefinitionQuery": new_vdq
        })
        self._track_view.last_known_locations_layer.manager.update_definition({
            "viewDefinitionQuery": new_vdq
        })
        # update cached values
        self._track_view.tracks_layer._hydrate()
        self._track_view.last_known_locations_layer._hydrate()

    @property
    def view_definition_query(self):
        """The View Definition Query of the tracks layer"""
        vdq = self._track_view.tracks_layer.properties["viewDefinitionQuery"]
        # Enterprise wraps VDQ in parentheses, but online doesn't
        if vdq[0] == "(" and vdq[-1] == ")":
            return vdq[1:-1]
        return vdq



"Functions for calling the Deep Learning Tools."

from arcgis.geoprocessing._support import _analysis_job, _analysis_job_results, \
     _analysis_job_status, _layer_input
import json as _json
import arcgis as _arcgis
from arcgis.raster._layer import ImageryLayer as _ImageryLayer
from arcgis.raster._util import _set_context, _id_generator
from .models import SingleShotDetector, UnetClassifier, FeatureClassifier, RetinaNet, PSPNetClassifier, EntityRecognizer, MaskRCNN
from ._data import prepare_data
from ._process_df import process_df, add_datepart

def _set_param(gis, params, param_name, input_param):
    if isinstance(input_param, str):
        if 'http:' in input_param or 'https:' in input_param:
            params[param_name] = _json.dumps({ 'url' : input_param })
        else:
            params[param_name] = _json.dumps({ 'uri' : input_param })

    elif isinstance(input_param, _arcgis.gis.Item):
        params[param_name] = _json.dumps({ "itemId" : input_param.itemid })

    elif isinstance(input_param, dict):
        params[param_name] =  input_param
    elif isinstance(input_param, Model):
        params[param_name] = input_param._model
    else:
        raise TypeError(input_param+" should be a string (service url) or Item")

    return


def _create_output_image_service(gis, output_name, task, folder = None):
    ok = gis.content.is_service_name_available(output_name, "Image Service")
    if not ok:
        raise RuntimeError("An Image Service by this name already exists: " + output_name)

    create_parameters = {
        "name": output_name,
        "description": "",
        "capabilities": "Image",
        "properties": {
            "path": "@",
            "description": "",
            "copyright": ""
        }
    }

    output_service = gis.content.create_service(output_name, create_params=create_parameters,
                                                service_type="imageService", folder = folder)
    description = "Image Service generated from running the " + task + " tool."
    item_properties = {
        "description": description,
        "tags": "Analysis Result, " + task,
        "snippet": "Analysis Image Service generated from " + task
    }
    output_service.update(item_properties)
    return output_service


def _create_output_feature_service(gis, output_name, output_service_name='Analysis feature service', task='GeoAnalytics', folder = None):
    ok = gis.content.is_service_name_available(output_name, 'Feature Service')
    if not ok:
        raise RuntimeError("A Feature Service by this name already exists: " + output_name)

    createParameters = {
        "currentVersion": 10.2,
            "serviceDescription": "",
            "hasVersionedData": False,
            "supportsDisconnectedEditing": False,
            "hasStaticData": True,
            "maxRecordCount": 2000,
            "supportedQueryFormats": "JSON",
            "capabilities": "Query",
            "description": "",
            "copyrightText": "",
            "allowGeometryUpdates": False,
            "syncEnabled": False,
            "editorTrackingInfo": {
                "enableEditorTracking": False,
                "enableOwnershipAccessControl": False,
                "allowOthersToUpdate": True,
                "allowOthersToDelete": True
                },
            "xssPreventionInfo": {
                "xssPreventionEnabled": True,
                "xssPreventionRule": "InputOnly",
                "xssInputRule": "rejectInvalid"
                },
            "tables": [],
            "name": output_service_name.replace(' ', '_')
    }

    output_service = gis.content.create_service(output_name, create_params=createParameters, service_type="featureService", folder =  folder)
    description = "Feature Service generated from running the " + task + " tool."
    item_properties = {
        "description" : description,
            "tags" : "Analysis Result, " + task,
            "snippet": output_service_name
    }
    output_service.update(item_properties)
    return output_service

def _set_output_raster(output_name, task, gis, output_properties =None):
    output_service = None
    output_raster = None
    
    task_name = task

    folder = None
    folderId = None

    if output_properties is not None:
        if "folder" in output_properties:
            folder = output_properties["folder"]
    if folder is not None:
        if isinstance(folder, dict):
            if "id" in folder:
                folderId = folder["id"]
                folder=folder["title"]
        else:
            owner = gis.properties.user.username
            folderId = gis._portal.get_folder_id(owner, folder)
        if folderId is None:
            folder_dict = gis.content.create_folder(folder, owner)
            folder = folder_dict["title"]
            folderId = folder_dict["id"]

    if output_name is None:
        output_name = str(task_name) + '_' + _id_generator()
        output_service = _create_output_image_service(gis, output_name, task, folder = folder)
        output_raster = {"serviceProperties": {"name" : output_service.name, "serviceUrl" : output_service.url}, "itemProperties": {"itemId" : output_service.itemid}}
    elif isinstance(output_name, str):
        output_service = _create_output_image_service(gis, output_name, task,folder = folder)
        output_raster = {"serviceProperties": {"name" : output_service.name, "serviceUrl" : output_service.url}, "itemProperties": {"itemId" : output_service.itemid}}
    elif isinstance(output_name, _arcgis.gis.Item):
        output_service = output_name
        output_raster = {"itemProperties":{"itemId":output_service.itemid}}
    else:
        raise TypeError("output_raster should be a string (service name) or Item") 

    if folderId is not None:
        output_raster["itemProperties"].update({"folderId":folderId})
    output_raster = _json.dumps(output_raster)
    return output_raster, output_service


def detect_objects(input_raster,
                   model,
                   model_arguments=None,
                   output_name=None,
                   run_nms=False,
                   confidence_score_field=None,
                   class_value_field=None,
                   max_overlap_ratio=0,
                   context=None,
                   process_all_raster_items=False,
                   *,
                   gis=None,
                   future=False,
                   **kwargs):

    """
    Function can be used to generate feature service that contains polygons on detected objects
    found in the imagery data using the designated deep learning model. Note that the deep learning
    library needs to be installed separately, in addition to the server's built in Python 3.x library.

    ====================================     ====================================================================
    **Argument**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    input_raster                             Required. raster layer that contains objects that needs to be detected.
    ------------------------------------     --------------------------------------------------------------------
    model                                    Required model object. 
    ------------------------------------     --------------------------------------------------------------------
    model_arguments                          Optional dictionary. Name-value pairs of arguments and their values that can be customized by the clients.
                                             
                                             eg: {"name1":"value1", "name2": "value2"}
    ------------------------------------     --------------------------------------------------------------------
    output_name                              Optional. If not provided, a Feature layer is created by the method and used as the output .
                                             You can pass in an existing Feature Service Item from your GIS to use that instead.
                                             Alternatively, you can pass in the name of the output Feature Service that should be created by this method
                                             to be used as the output for the tool.
                                             A RuntimeError is raised if a service by that name already exists
    ------------------------------------     --------------------------------------------------------------------
    run_nms                                  Optional bool. Default value is False. If set to True, runs the Non Maximum Suppression tool.
    ------------------------------------     --------------------------------------------------------------------
    confidence_score_field                   Optional string. The field in the feature class that contains the confidence scores as output by the object detection method.
                                             This parameter is required when you set the run_nms to True
    ------------------------------------     --------------------------------------------------------------------
    class_value_field                        Optional string. The class value field in the input feature class. 
                                             If not specified, the function will use the standard class value fields 
                                             Classvalue and Value. If these fields do not exist, all features will 
                                             be treated as the same object class.
                                             Set only if run_nms  is set to True
    ------------------------------------     --------------------------------------------------------------------
    max_overlap_ratio                        Optional integer. The maximum overlap ratio for two overlapping features. 
                                             Defined as the ratio of intersection area over union area. 
                                             Set only if run_nms  is set to True
    ------------------------------------     --------------------------------------------------------------------
    context                                  Optional dictionary. Context contains additional settings that affect task execution.
                                             Dictionary can contain value for following keys:

                                             - cellSize - Set the output raster cell size, or resolution

                                             - extent - Sets the processing extent used by the function

                                             - parallelProcessingFactor - Sets the parallel processing factor. Default is "80%"

                                             - processorType - Sets the processor type. "CPU" or "GPU"

                                             Eg: {"processorType" : "CPU"}

                                             Setting context parameter will override the values set using arcgis.env 
                                             variable for this particular function.
    ------------------------------------     --------------------------------------------------------------------
    process_all_raster_items                 Optional bool. Specifies how all raster items in an image service will be processed.

                                              - False : all raster items in the image service will be mosaicked together and processed. This is the default.

                                              - True : all raster items in the image service will be processed as separate images.
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ------------------------------------     --------------------------------------------------------------------
    future                                   Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ====================================     ====================================================================

    :return:
        The output feature layer item containing the detected objects

    """


    #task = "DetectObjectsUsingDeepLearning"

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.rasteranalysis.detect_objects_using_deep_learning(input_raster=input_raster,
                                                                        model=model,
                                                                        output_objects=output_name,
                                                                        model_arguments=model_arguments,
                                                                        run_nms=run_nms,
                                                                        confidence_score_field=confidence_score_field,
                                                                        class_value_field=class_value_field,
                                                                        max_overlap_ratio=max_overlap_ratio,
                                                                        context=context,
                                                                        process_all_raster_items=process_all_raster_items,
                                                                        future=future,
                                                                        **kwargs)

    """
    url = gis.properties.helperServices.rasterAnalytics.url
    gptool = _arcgis.gis._GISResource(url, gis)

    params = {}

    params["inputRaster"] = _layer_input(input_raster)

    if output_name is None:
        output_service_name = 'DetectObjectsUsingDeepLearning_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    folder = None
    folderId = None
    if kwargs is not None:
        if "folder" in kwargs:
                folder = kwargs["folder"]
        if folder is not None:
            if isinstance(folder, dict):
                if "id" in folder:
                    folderId = folder["id"]
                    folder=folder["title"]
            else:
                owner = gis.properties.user.username
                folderId = gis._portal.get_folder_id(owner, folder)
            if folderId is None:
                folder_dict = gis.content.create_folder(folder, owner)
                folder = folder_dict["title"]
                folderId = folder_dict["id"]

    output_service = _create_output_feature_service(gis, output_name, output_service_name, 'Detect Objects', folder)

    if folderId is not None:
        params["outputObjects"] = _json.dumps({"serviceProperties": {"name": output_service_name, "serviceUrl": output_service.url},
                                               "itemProperties": {"itemId": output_service.itemid}, "folderId":folderId})
    else:
        params["outputObjects"] = _json.dumps({"serviceProperties": {"name": output_service_name, "serviceUrl": output_service.url},
                                               "itemProperties": {"itemId": output_service.itemid}})

    if model is None:
        raise RuntimeError('model cannot be None')
    else:
        _set_param(gis, params, "model", model)

    if model_arguments:
        params["modelArguments"] = dict((str(k),str(v)) for k, v in model_arguments.items())

    if isinstance(run_nms, bool):
        if run_nms:
            params["runNMS"] = True

            if confidence_score_field is not None:
                params["confidenceScoreField"] = confidence_score_field

            if class_value_field is not None:
                params["classValueField"] = class_value_field
    
            if max_overlap_ratio is not None:
                params["maxOverlapRatio"] = max_overlap_ratio
        else:
            params["runNMS"] = False
    else:
        raise RuntimeError("run_nms value should be an instance of bool")
    
    _set_context(params, context)

    task_url, job_info, job_id = _analysis_job(gptool, task, params)

    job_info = _analysis_job_status(gptool, task_url, job_info)
    job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
    item_properties = {
        "properties": {
            "jobUrl": task_url + '/jobs/' + job_info['jobId'],
            "jobType": "GPServer",
            "jobId": job_info['jobId'],
            "jobStatus": "completed"
        }
    }
    output_service.update(item_properties)
    return output_service
    """

def classify_pixels(input_raster,
                    model,
                    model_arguments=None,
                    output_name=None,
                    context=None,
                    process_all_raster_items=False,
                    *,
                    gis=None,
                    future=False,
                    **kwargs):

    """
    Function to classify input imagery data using a deep learning model.
    Note that the deep learning library needs to be installed separately,
    in addition to the server's built in Python 3.x library.

    ====================================     ====================================================================
    **Argument**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    input_raster                             Required. raster layer that needs to be classified
    ------------------------------------     --------------------------------------------------------------------
    model                                    Required model object.
    ------------------------------------     --------------------------------------------------------------------
    model_arguments                          Optional dictionary. Name-value pairs of arguments and their values that can be customized by the clients.

                                             eg: {"name1":"value1", "name2": "value2"}

    ------------------------------------     --------------------------------------------------------------------
    output_name                              Optional. If not provided, an imagery layer is created by the method and used as the output .
                                             You can pass in an existing Image Service Item from your GIS to use that instead.
                                             Alternatively, you can pass in the name of the output Image Service that should be created by this method
                                             to be used as the output for the tool.
                                             A RuntimeError is raised if a service by that name already exists
    ------------------------------------     --------------------------------------------------------------------
    context                                  Optional dictionary. Context contains additional settings that affect task execution.
                                               Dictionary can contain value for following keys:

                                               - outSR - (Output Spatial Reference) Saves the result in the specified spatial reference

                                               - snapRaster - Function will adjust the extent of output rasters so that they 
                                                 match the cell alignment of the specified snap raster.

                                               - cellSize - Set the output raster cell size, or resolution

                                               - extent - Sets the processing extent used by the function

                                               - parallelProcessingFactor - Sets the parallel processing factor. Default is "80%"

                                               - processorType - Sets the processor type. "CPU" or "GPU"

                                               Eg: {"outSR" : {spatial reference}}

                                               Setting context parameter will override the values set using arcgis.env 
                                               variable for this particular function.
    ------------------------------------     --------------------------------------------------------------------
    process_all_raster_items                 Optional bool. Specifies how all raster items in an image service will be processed.

                                              - False : all raster items in the image service will be mosaicked together and processed. This is the default.

                                              - True : all raster items in the image service will be processed as separate images.
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ------------------------------------     --------------------------------------------------------------------
    future                                   Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ====================================     ====================================================================

    :return:
        The classified imagery layer item

    """


    #task = "ClassifyPixelsUsingDeepLearning"

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.rasteranalysis.classify_pixels_using_deep_learning(input_raster=input_raster,
                                                                        model=model,
                                                                        model_arguments=model_arguments,
                                                                        output_classified_raster=output_name,
                                                                        context=context,
                                                                        process_all_raster_items=process_all_raster_items,
                                                                        future=future,
                                                                        **kwargs)


    """
    url = gis.properties.helperServices.rasterAnalytics.url
    gptool = _arcgis.gis._GISResource(url, gis)

    output_service = None

    output_raster, output_service = _set_output_raster(output_name, task, gis, kwargs)

    params = {}

    params["outputClassifiedRaster"] = output_raster

    params["inputRaster"] = _layer_input(input_raster)

    if model is None:
        raise RuntimeError('model cannot be None')
    else:
        _set_param(gis, params, "model", model)

    if model_arguments:
        params["modelArguments"] = dict((str(k),str(v)) for k, v in model_arguments.items())

    _set_context(params, context)

    task_url, job_info, job_id = _analysis_job(gptool, task, params)

    job_info = _analysis_job_status(gptool, task_url, job_info)
    job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
    item_properties = {
        "properties": {
            "jobUrl": task_url + '/jobs/' + job_info['jobId'],
            "jobType": "GPServer",
            "jobId": job_info['jobId'],
            "jobStatus": "completed"
        }
    }
    output_service.update(item_properties)
    return output_service
    """

def export_training_data(input_raster,
                         input_class_data=None,
                         chip_format=None,
                         tile_size=None,
                         stride_size=None,
                         metadata_format=None,
                         classvalue_field=None,
                         buffer_radius=None,
                         output_location=None,
                         context=None,
                         input_mask_polygons=None,
                         rotation_angle=0,
                         reference_system="MAP_SPACE",
                         process_all_raster_items=False,
                         blacken_around_feature=False,
                         fix_chip_size=True,
                         *,
                         gis=None,
                         future=False,
                         **kwargs):

    """
    Function is designed to generate training sample image chips from the input imagery data with
    labeled vector data or classified images. The output of this service tool is the data store string
    where the output image chips, labels and metadata files are going to be stored.

    ====================================     ====================================================================
    **Argument**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    input_raster                             Required. Raster layer that needs to be exported for training
    ------------------------------------     --------------------------------------------------------------------
    input_class_data                         Labeled data, either a feature layer or image layer.
                                             Vector inputs should follow a training sample format as
                                             generated by the ArcGIS Pro Training Sample Manager.
                                             Raster inputs should follow a classified raster format as generated by the Classify Raster tool.
    ------------------------------------     --------------------------------------------------------------------
    chip_format                              Optional string. The raster format for the image chip outputs.

                                                - TIFF: TIFF format

                                                - PNG: PNG format

                                                - JPEG: JPEG format

                                                - MRF: MRF (Meta Raster Format)
    ------------------------------------     --------------------------------------------------------------------
    tile_size                                Optional dictionary. The size of the image chips.

                                                Example: {"x": 256, "y": 256}
    ------------------------------------     --------------------------------------------------------------------
    stride_size                              Optional dictionary. The distance to move in the X and Y when creating 
                                             the next image chip.
                                             When stride is equal to the tile size, there will be no overlap.
                                             When stride is equal to half of the tile size, there will be 50% overlap.

                                               Example: {"x": 128, "y": 128}
    ------------------------------------     --------------------------------------------------------------------
    metadata_format                          Optional string. The format of the output metadata labels. There are 4 options for output metadata labels for the training data,
                                               KITTI Rectangles, PASCAL VOCrectangles, Classified Tiles (a class map) and RCNN_Masks. If your input training sample data
                                               is a feature class layer such as building layer or standard classification training sample file,
                                               use the KITTI or PASCAL VOC rectangle option.

                                               The output metadata is a .txt file or .xml file containing the training sample data contained
                                               in the minimum bounding rectangle. The name of the metadata file matches the input source image
                                               name. If your input training sample data is a class map, use the Classified Tiles as your output metadata format option.

                                               - KITTI_rectangles: The metadata follows the same format as the Karlsruhe Institute of Technology and Toyota
                                                 Technological Institute (KITTI) Object Detection Evaluation dataset. The KITTI dataset is a vision benchmark suite.
                                                 This is the default.The label files are plain text files. All values, both numerical or strings, are separated by
                                                 spaces, and each row corresponds to one object.

                                               - PASCAL_VOC_rectangles: The metadata follows the same format as the Pattern Analysis, Statistical Modeling and
                                                 Computational Learning, Visual Object Classes (PASCAL_VOC) dataset. The PASCAL VOC dataset is a standardized
                                                 image data set for object class recognition.The label files are XML files and contain information about image name,
                                                 class value, and bounding box(es).

                                               - Classified_Tiles: This option will output one classified image chip per input image chip.
                                                 No other meta data for each image chip. Only the statistics output has more information on the
                                                 classes such as class names, class values, and output statistics.

                                               - RCNN_Masks: This option will output image chips that have a mask on the areas where the sample exists.
                                                 The model generates bounding boxes and segmentation masks for each instance of an object in the image.
                                                 It's based on Feature Pyramid Network (FPN) and a ResNet101 backbone.

                                               - Labeled_Tiles : This option will label each output tile with a specific class.
    ------------------------------------     --------------------------------------------------------------------
    classvalue_field                          Optional string. Specifies the field which contains the class values. If no field is specified,
                                              the system will look for a 'value' or 'classvalue' field. If this feature does
                                              not contain a class field, the system will presume all records belong the 1 class.
    ------------------------------------     --------------------------------------------------------------------
    buffer_radius                            Optional integer. Specifies a radius for point feature classes to specify training sample area.
    ------------------------------------     --------------------------------------------------------------------
    output_location                          This is the output location for training sample data.
                                               It can be the server data store path or a shared file system path.

                                               Example:

                                               Server datastore path -
                                                ``/fileShares/deeplearning/rooftoptrainingsamples``
                                                ``/rasterStores/rasterstorename/rooftoptrainingsamples``
                                                ``/cloudStores/cloudstorename/rooftoptrainingsamples``

                                               File share path - 
                                                ``\\\\servername\\deeplearning\\rooftoptrainingsamples``
    ------------------------------------     --------------------------------------------------------------------
    context                                  Optional dictionary. Context contains additional settings that affect task execution.
                                                Dictionary can contain value for following keys:

                                                - exportAllTiles - Choose if the image chips with overlapped labeled data will be exported.
                                                    True - Export all the image chips, including those that do not overlap labeled data. 
                                                    False - Export only the image chips that overlap the labelled data. This is the default.

                                                - startIndex - Allows you to set the start index for the sequence of image chips.
                                                    This lets you append more image chips to an existing sequence. The default value is 0.

                                                - cellSize - cell size can be set using this key in context parameter

                                                - extent - Sets the processing extent used by the function

                                                Setting context parameter will override the values set using arcgis.env 
                                                variable for this particular function.(cellSize, extent)

                                                eg: {"exportAllTiles" : False, "startIndex": 0 }
    ------------------------------------     --------------------------------------------------------------------
    input_mask_polygons                       Optional feature layer. The feature layer that delineates the area where 
                                               image chips will be created.
                                               Only image chips that fall completely within the polygons will be created.
    ------------------------------------     --------------------------------------------------------------------
    rotation_angle                           Optional float. The rotation angle that will be used to generate additional 
                                               image chips.

                                               An image chip will be generated with a rotation angle of 0, which 
                                               means no rotation. It will then be rotated at the specified angle to 
                                               create an additional image chip. The same training samples will be 
                                               captured at multiple angles in multiple image chips for data augmentation.
                                               The default rotation angle is 0.
    ------------------------------------     --------------------------------------------------------------------
    reference_system                         Optional string. Specifies the type of reference system to be used to interpret 
                                             the input image. The reference system specified should match the reference system 
                                             used to train the deep learning model. 

                                             - MAP_SPACE : The input image is in a map-based coordinate system. This is the default.

                                             - IMAGE_SPACE : The input image is in image space, viewed from the direction of the sensor 
                                               that captured the image, and rotated such that the tops of buildings and trees point upward in the image.

                                             - PIXEL_SPACE : The input image is in image space, with no rotation and no distortion. 
    ------------------------------------     --------------------------------------------------------------------
    process_all_raster_items                 Optional bool. Specifies how all raster items in an image service will be processed.

                                              - False : all raster items in the image service will be mosaicked together and processed. This is the default.

                                              - True : all raster items in the image service will be processed as separate images.
    ------------------------------------     --------------------------------------------------------------------
    blacken_around_feature                   Optional bool. 
                                             
                                             Specifies whether to blacken the pixels around each object or feature in each image tile.

                                             This parameter only applies when the metadata format is set to Labeled_Tiles and an input feature class or classified raster has been specified.

                                             - False : Pixels surrounding objects or features will not be blackened. This is the default.

                                             - True : Pixels surrounding objects or features will be blackened.

    ------------------------------------     --------------------------------------------------------------------
    fix_chip_size                            Optional bool. Specifies whether to crop the exported tiles such that they are all the same size.

                                             This parameter only applies when the metadata format is set to Labeled_Tiles and an input feature class or classified raster has been specified.

                                             - True : Exported tiles will be the same size and will center on the feature. This is the default.

                                             - False : Exported tiles will be cropped such that the bounding geometry surrounds only the feature in the tile.
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ------------------------------------     --------------------------------------------------------------------
    future                                   Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ====================================     ====================================================================

    :return:
        Output string containing the location of the exported training data

    """

    #task = "ExportTrainingDataforDeepLearning"

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.rasteranalysis.export_training_data_for_deep_learning(input_raster=input_raster,
                                                                            input_class_data=input_class_data,
                                                                            chip_format=chip_format,
                                                                            tile_size=tile_size,
                                                                            stride_size=stride_size,
                                                                            metadata_format=metadata_format,
                                                                            class_value_field=classvalue_field,
                                                                            buffer_radius=buffer_radius,
                                                                            output_location=output_location,
                                                                            input_mask_polygons=input_mask_polygons,
                                                                            rotation_angle=rotation_angle,
                                                                            reference_system=reference_system,
                                                                            process_all_raster_items=process_all_raster_items,
                                                                            blacken_around_feature=blacken_around_feature,
                                                                            fix_chip_size=fix_chip_size,
                                                                            context=context,
                                                                            future=future,
                                                                            **kwargs)

    """
    url = gis.properties.helperServices.rasterAnalytics.url
    gptool = _arcgis.gis._GISResource(url, gis)

    params = {}

    if output_location:
        params["outputLocation"] = output_location
    else:
        raise RuntimeError("output_location cannot be None")

    if input_raster:
        params["inputRaster"] = _layer_input(input_raster)
    else:
        raise RuntimeError("input_raster cannot be None")

    if input_class_data:
        params["inputClassData"] = _layer_input(input_class_data)

    if chip_format is not None:
        chipFormatAllowedValues = ['TIFF', 'PNG', 'JPEG','MRF']
        if not chip_format in chipFormatAllowedValues:
            raise RuntimeError('chip_format can only be one of the following: '+ str(chipFormatAllowedValues))
        params["chipFormat"] = chip_format

    if tile_size:
        params["tileSize"] = tile_size

    if stride_size:
        params["strideSize"] = stride_size

    if metadata_format is not None:
        metadataFormatAllowedValues = ['KITTI_rectangles', 'PASCAL_VOC_rectangles', 'Classified_Tiles', 'RCNN_Masks', 'Labeled_Tiles']
        if not metadata_format in metadataFormatAllowedValues:
            raise RuntimeError('metadata_format can only be one of the following: '+ str(metadataFormatAllowedValues))

        params['metadataFormat'] = metadata_format

    if buffer_radius is not None:
        params["bufferRadius"]= buffer_radius

    if classvalue_field is not None:
        params["classValueField"]= classvalue_field

    if input_mask_polygons:
        params["inputMaskPolygons"] = _layer_input(input_mask_polygons)

    if rotation_angle:
        params["rotationAngle"] = rotation_angle

    _set_context(params, context)

    task_url, job_info, job_id = _analysis_job(gptool, task, params)

    job_info = _analysis_job_status(gptool, task_url, job_info)
    job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
    item_properties = {
        "properties": {
            "jobUrl": task_url + '/jobs/' + job_info['jobId'],
            "jobType": "GPServer",
            "jobId": job_info['jobId'],
            "jobStatus": "completed"
        }
    }
    return job_values["outLocation"]["uri"]

    """


def list_models(*,
                gis=None,
                future=False,
                **kwargs):
    """
    Function is used to list all the installed deep learning models.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ------------------     --------------------------------------------------------------------
    future                 Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ==================     ====================================================================

    :return:
        list of deep learning models installed

    """



    #task = "ListDeepLearningModels"

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.rasteranalysis.list_deep_learning_models(future=future,
                                                               **kwargs)
    """
    url = gis.properties.helperServices.rasterAnalytics.url
    gptool = _arcgis.gis._GISResource(url, gis)
    params = {}
    task_url, job_info, job_id = _analysis_job(gptool, task, params)

    job_info = _analysis_job_status(gptool, task_url, job_info)
    job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
    item_properties = {
        "properties": {
            "jobUrl": task_url + '/jobs/' + job_info['jobId'],
            "jobType": "GPServer",
            "jobId": job_info['jobId'],
            "jobStatus": "completed"
        }
    }

    output_model_list = []
    if isinstance(job_values["deepLearningModels"], list) and job_values["deepLearningModels"] is not None:
        for element in job_values["deepLearningModels"]:
            if isinstance(element,dict):
                if "id" in element.keys():
                    item = gis.content.get(element["id"])
                    output_model_list.append(Model(item))
    return output_model_list
    """

def classify_objects(input_raster,
                     model,
                     model_arguments=None,
                     input_features=None,
                     class_label_field=None,
                     process_all_raster_items=False,
                     output_name=None,
                     context=None,
                     *,
                     gis=None,
                     future=False,
                     **kwargs):

    """
    Function can be used to output feature service with assigned class label for each feature based on
    information from overlapped imagery data using the designated deep learning model. 

    ====================================     ====================================================================
    **Argument**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    input_raster                             Required. raster layer that contains objects that needs to be classified.
    ------------------------------------     --------------------------------------------------------------------
    model                                    Required model object.
    ------------------------------------     --------------------------------------------------------------------
    model_arguments                          Optional dictionary. Name-value pairs of arguments and their values that can be customized by the clients.
                                             
                                             eg: {"name1":"value1", "name2": "value2"}
    ------------------------------------     --------------------------------------------------------------------
    input_features                           Optional feature layer.
                                             The point, line, or polygon input feature layer that identifies the location of each object to be 
                                             classified and labelled. Each row in the input feature layer represents a single object.

                                             If no input feature layer is specified, the function assumes that each input image contains a single object 
                                             to be classified. If the input image or images use a spatial reference, the output from the function is a 
                                             feature layer, where the extent of each image is used as the bounding geometry for each labelled 
                                             feature layer. If the input image or images are not spatially referenced, the output from the function 
                                             is a table containing the image ID values and the class labels for each image.
    ------------------------------------     --------------------------------------------------------------------
    class_label_field                        Optional str. The name of the field that will contain the classification label in the output feature layer.

                                             If no field name is specified, a new field called ClassLabel will be generated in the output feature layer.

                                             Example:
                                                "ClassLabel"
    ------------------------------------     --------------------------------------------------------------------
    process_all_raster_items                 Optional bool. 

                                             If set to False, all raster items in the image service will be mosaicked together and processed. This is the default.

                                             If set to True, all raster items in the image service will be processed as separate images.
    ------------------------------------     --------------------------------------------------------------------
    output_name                              Optional. If not provided, a Feature layer is created by the method and used as the output .
                                             You can pass in an existing Feature Service Item from your GIS to use that instead.
                                             Alternatively, you can pass in the name of the output Feature Service that should be created by this method
                                             to be used as the output for the tool.
                                             A RuntimeError is raised if a service by that name already exists
    ------------------------------------     --------------------------------------------------------------------
    context                                  Optional dictionary. Context contains additional settings that affect task execution.
                                             Dictionary can contain value for following keys:

                                             - cellSize - Set the output raster cell size, or resolution

                                             - extent - Sets the processing extent used by the function

                                             - parallelProcessingFactor - Sets the parallel processing factor. Default is "80%"

                                             - processorType - Sets the processor type. "CPU" or "GPU"

                                             Eg: {"processorType" : "CPU"}

                                             Setting context parameter will override the values set using arcgis.env 
                                             variable for this particular function.
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    ====================================     ====================================================================

    :return:
        The output feature layer item containing the classified objects

    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.rasteranalysis.classify_objects_using_deep_learning(input_raster=input_raster,
                                                                          input_features=input_features,
                                                                          output_feature_class=output_name,
                                                                          model=model,
                                                                          model_arguments=model_arguments,
                                                                          class_label_field=class_label_field,
                                                                          process_all_raster_items=process_all_raster_items,
                                                                          context=context,
                                                                          future=future,
                                                                          **kwargs)

class Model:
    def __init__(self, model = None):
        self._model_package = False
        if isinstance(model, _arcgis.gis.Item):
            self._model = _json.dumps({ "itemId" : model.itemid })
            self._model_package = True
            self.item = model

    def _repr_html_(self):
        if self._model_package:
            if  hasattr(self,"item"):
                self.item._repr_html_()
        else:
            self.__repr__()

    def __repr__(self):
        if self._model_package:
            model = _json.loads(self._model)
            if "url" in model.keys():
                return '<Model:%s>' % self._model
            return '<Model Title:%s owner:%s>' %(self.item.title, self.item.owner)

        else:
            try:
                return '<Model:%s>' % self._model
            except:
                return '<empty Model>'


    def from_json(self, model):
        """
        Function is used to initialise Model object from model definition JSON
        
        eg usage:

        model = Model()

        model.from_json({"Framework" :"TensorFlow",
                        "ModelConfiguration":"DeepLab",
                        "InferenceFunction":"``[functions]System\\DeepLearning\\ImageClassifier.py``",
                        "ModelFile":"``\\\\folder_path_of_pb_file\\frozen_inference_graph.pb``",
                        "ExtractBands":[0,1,2],
                        "ImageWidth":513,
                        "ImageHeight":513,
                        "Classes": [ { "Value":0, "Name":"Evergreen Forest", "Color":[0, 51, 0] },
                                     { "Value":1, "Name":"Grassland/Herbaceous", "Color":[241, 185, 137] },
                                     { "Value":2, "Name":"Bare Land", "Color":[236, 236, 0] },
                                     { "Value":3, "Name":"Open Water", "Color":[0, 0, 117] },
                                     { "Value":4, "Name":"Scrub/Shrub", "Color":[102, 102, 0] },
                                     { "Value":5, "Name":"Impervious Surface", "Color":[236, 236, 236] } ] })

        """
        if isinstance(model, dict):
            self._model = model
            self._model_package = False

    def from_model_path(self, model):
        """
        Function is used to initialise Model object from url of model package or path of model definition file
        eg usage:

        model = Model()

        model.from_model_path("https://xxxportal.esri.com/sharing/rest/content/items/<itemId>")

        or
        model = Model()

        model.from_model_path(``"\\\\sharedstorage\\sharefolder\\findtrees.emd"``)

        """
        if 'http:' in model or 'https:' in model:
            self._model = _json.dumps({ 'url' : model })
            self._model_package = True
        else:
            self._model = _json.dumps({ 'uri' : model })
            self._model_package = False


    def install(self,
                *,
                gis=None,
                future=False,
                **kwargs):

        """
        Function is used to install the uploaded model package (*.dlpk). Optionally after inferencing
        the necessary information using the model, the model can be uninstalled by uninstall_model()


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
        ------------------     --------------------------------------------------------------------
        future                 Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
        ==================     ====================================================================

        :return:
            Path where model is installed

        """
        if self._model_package is False:
            raise RuntimeError("model object should be created from a portal item or a portal url")

        if self._model is None:
            raise RuntimeError("For install/uninstall model object should be created from a portal item or portal url")

        #task = "InstallDeepLearningModel"

        gis = _arcgis.env.active_gis if gis is None else gis
        return gis._tools.rasteranalysis.install_deep_learning_model(model_package=self._model,
                                                                     future=future,
                                                                     **kwargs)

        """
        url = gis.properties.helperServices.rasterAnalytics.url
        gptool = _arcgis.gis._GISResource(url, gis)

        params = {}

        if self._model is None:
            raise RuntimeError("For install/uninstall model object should be created from a portal item or portal url")
        else:
            params["modelPackage"] = self._model

        task_url, job_info, job_id = _analysis_job(gptool, task, params)

        job_info = _analysis_job_status(gptool, task_url, job_info)
        job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
        item_properties = {
            "properties": {
                "jobUrl": task_url + '/jobs/' + job_info['jobId'],
                "jobType": "GPServer",
                "jobId": job_info['jobId'],
                "jobStatus": "completed"
            }
        }

        return job_values["installSucceed"]
        """


    def query_info(self,
                   *,
                   gis=None,
                   future=False,
                   **kwargs):
        """
        Function is used to extract the deep learning model specific settings from the model package item or model definition file.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
        ------------------     --------------------------------------------------------------------
        future                 Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
        ==================     ====================================================================

        :return:
           The key model information in dictionary format that describes what the settings are essential for this type of deep learning model.
        """


        #task = "QueryDeepLearningModelInfo"

        gis = _arcgis.env.active_gis if gis is None else gis
        if self._model is None:
            raise RuntimeError('model cannot be None')

        return gis._tools.rasteranalysis.query_deep_learning_model_info(model=self._model,
                                                                        future=future,
                                                                        **kwargs)

        """
        url = gis.properties.helperServices.rasterAnalytics.url
        gptool = _arcgis.gis._GISResource(url, gis)

        params = {}

        if self._model is None:
            raise RuntimeError('model cannot be None')
        else:
            params["model"] = self._model

        task_url, job_info, job_id = _analysis_job(gptool, task, params)

        job_info = _analysis_job_status(gptool, task_url, job_info)
        job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
        item_properties = {
            "properties": {
                "jobUrl": task_url + '/jobs/' + job_info['jobId'],
                "jobType": "GPServer",
                "jobId": job_info['jobId'],
                "jobStatus": "completed"
            }
        }
        output = job_values["outModelInfo"]
        print(output)
        try:
            dict_output =  _json.loads(output["modelInfo"])
            return dict_output
        except:
            return output
        """

    def uninstall(self,
                  *,
                  gis=None,
                  future=False,
                  **kwargs):

        """
        Function is used to uninstall the uploaded model package that was installed using the install_model()
        This function will delete the named deep learning model from the server but not the portal item.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        gis                    Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
        ------------------     --------------------------------------------------------------------
        future                 Keyword only parameter. Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
        ==================     ====================================================================

        :return:
            itemId of the uninstalled model package item

        """
        if self._model_package is False:
            raise RuntimeError("For install/uninstall model object should be created from a portal item or a portal url")

        #task = "UninstallDeepLearningModel"

        gis = _arcgis.env.active_gis if gis is None else gis

        if self._model is None:
            raise RuntimeError('model_package cannot be None')

        return gis._tools.rasteranalysis.uninstall_deep_learning_model(model_item_id=self._model,
                                                                       future=future,
                                                                       **kwargs)

        """
        url = gis.properties.helperServices.rasterAnalytics.url
        gptool = _arcgis.gis._GISResource(url, gis)

        params = {}

        if self._model is None:
            raise RuntimeError('model_package cannot be None')
        else:
            params["modelItemId"] = self._model

        task_url, job_info, job_id = _analysis_job(gptool, task, params)

        job_info = _analysis_job_status(gptool, task_url, job_info)
        job_values = _analysis_job_results(gptool, task_url, job_info, job_id)
        item_properties = {
            "properties": {
                "jobUrl": task_url + '/jobs/' + job_info['jobId'],
                "jobType": "GPServer",
                "jobId": job_info['jobId'],
                "jobStatus": "completed"
            }
        }

        return job_values["uninstallSucceed"]
        """



"""
These functions help you explore the character of areas. Detailed demographic data and statistics are returned for your
chosen areas.

enrich_layer retrieves information about the people, places, and businesses in a specific area, or within a selected
travel time or distance from a location.
"""

import arcgis as _arcgis
import arcgis.network as network

def enrich_layer(input_layer,
                 data_collections=[],
                 analysis_variables=[],
                 country=None,
                 buffer_type=None,
                 distance=None,
                 units=None,
                 output_name=None,
                 context=None,
                 gis=None,
                 estimate=False,
                 return_boundaries=False,
                 future=False):
    """
    .. image:: _static/images/enrich_layer/enrich_layer.png

    The ``enrich_layer`` method enriches your data by getting facts about the people, places, and businesses that surround
    your data locations. For example: What kind of people live here? What do people like to do in this area? What are
    their habits and lifestyles? What kind of businesses are there in this area?

    The result will be a new layer of input features that includes all demographic and geographic information from given data collections.

    =====================================================================     ====================================================================
    **Parameter**                                                             **Description**
    ---------------------------------------------------------------------     --------------------------------------------------------------------
    input_layer                                                               Required layer. The features to enrich with new data. See :ref:`Feature Input<FeatureInput>`.
    ---------------------------------------------------------------------     --------------------------------------------------------------------
    data_collections                                                          Optional list of strings. This optional parameter defines the collections of data you want to use to enrich your features.
                                                                              Its value is a list of strings. If you don't provide this parameter, you must provide the analysis_variables parameter.

                                                                              For more information about data collections and the values for this parameter, visit the `Esri Demographics site <http://doc.arcgis.com/en/esri-demographics/data/data-browser.htm>`_.
    ---------------------------------------------------------------------     --------------------------------------------------------------------
    analysis_variables                                                        Optional list of strings. The parameter defines the specific variables within a data collection you want to use to
                                                                              your features. Its value is a list of strings in the form of "dataCollection.VariableName". If you don't provide
                                                                              this parameter, you must provide the dataCollections parameter. You can provide both parameters.
                                                                              For example, if you want all variables in the KeyGlobalFacts data collection, specify it in the dataCollections
                                                                              parameter and use this parameter for specific variables in other collections.

                                                                              For more information about variables in data collections, visit the `Esri Demographics site <http://doc.arcgis.com/en/esri-demographics/data/data-browser.htm>`_. Each data collection
                                                                              has a PDF file describing variables and their names.
    ---------------------------------------------------------------------     --------------------------------------------------------------------
    country                                                                   Optional string. This optional parameter further defines what is returned from data collection.
                                                                              For example, your input features may be countries in Western Europe, and you want to enrich them with the
                                                                              KeyWEFacts data collection. However, you only want data for France, not every country in your input layer.
                                                                              The value is the two-character country code.

                                                                              For more information about data collections and the values for this parameter, visit the `Esri Demographics site <http://doc.arcgis.com/en/esri-demographics/data/data-browser.htm>`_.
    ---------------------------------------------------------------------     --------------------------------------------------------------------
    buffer_type (Required if input_layer contains point or line features)     Optional string. If your input features are points or lines, you must define an area around your features that you want to enrich. Features that are within (or equal to) the distances you enter will be enriched.

                                                                              Choice list: ['StraightLine', 'Driving Distance', 'Driving Time ', 'Rural Driving Distance', 'Rural Driving Time', 'Trucking Distance', 'Trucking Time', 'Walking Distance', 'Walking Time']
    ---------------------------------------------------------------------     --------------------------------------------------------------------
    distance (Required if input_layer contains point or line features)        Optional float. A value that defines the search distance or time. The units of the distance value is supplied by the units parameter.
    ---------------------------------------------------------------------     --------------------------------------------------------------------
    units                                                                     Optional string. The linear unit to be used with the distance value(s) specified in the distance parameter.

                                                                              Choice list: ['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles', 'Seconds', 'Minutes'. 'Hours']
    ---------------------------------------------------------------------     --------------------------------------------------------------------
    output_name                                                               Optional string. If provided, the task will create a feature service of the results.
                                                                              You define the name of the service. If output_name is not supplied, the task will return a feature collection.
    ---------------------------------------------------------------------     --------------------------------------------------------------------
    context                                                                   Optional string. Context contains additional settings that affect task execution. For ``enrich_layer`` method, there are two settings.

                                                                              #. Extent (extent)-a bounding box that defines the analysis area. Only those features in the input_layer that intersect the bounding box will be analyzed.
                                                                              #. Output Spatial Reference (outSR) the output features will be projected into the output spatial reference.
    ---------------------------------------------------------------------     --------------------------------------------------------------------
    gis                                                                       Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    ---------------------------------------------------------------------     --------------------------------------------------------------------
    estimate                                                                  Optional Boolean. If True, the number of credits to run the operation will be returned.
    ---------------------------------------------------------------------     --------------------------------------------------------------------
    return_boundaries                                                         Optional boolean. Applies only for point and line input features. If True, a result layer of areas is returned.
                                                                              The returned areas are defined by the specified buffer_type. For example, if using a buffer_type of StraightLine with
                                                                              a distance of 5 miles, your result will contain areas with a 5 mile radius around the input features and requested
                                                                              analysis_variables variables. If False, the resulting layer will return the same features as the input layer with
                                                                              analysis_variables variables.

                                                                              The default value is False.
    ---------------------------------------------------------------------     --------------------------------------------------------------------
    future                                                                    Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    =====================================================================     ====================================================================

    :returns result_layer : feature layer Item if output_name is specified, else Feature Collection.

    .. code-block:: python

        USAGE EXAMPLE: To enrich US block groups with population as analysis variable.
        blkgrp_enrich = enrich_layer(block_groups,
                                     analysis_variables=["AtRisk.MP27002A_B"],
                                     country='US',
                                     output_name='enrich layer')

    """

    gis = _arcgis.env.active_gis if gis is None else gis
    if isinstance(buffer_type, str):
        if buffer_type != 'StraightLine':
            route_service = network.RouteLayer(gis.properties.helperServices.route.url, gis=gis)
            buffer_type = [i for i in route_service.retrieve_travel_modes()['supportedTravelModes'] if i['name'] == buffer_type][0]
        else:
            pass

    return gis._tools.featureanalysis.enrich_layer(
                 input_layer,
                 data_collections,
                 analysis_variables,
                 country,
                 buffer_type,
                 distance,
                 units,
                 output_name,
                 context,
                 estimate=estimate,
                 return_boundaries=return_boundaries,future=future)



try:
    import pandas as pd
    import numpy as np
    import re
    import sklearn
    import datetime
    from sklearn_pandas import DataFrameMapper
    from sklearn.preprocessing import StandardScaler
    from pandas.api.types import is_string_dtype, is_numeric_dtype
    import logging
    from sklearn.model_selection import StratifiedShuffleSplit
    from sklearn.model_selection import train_test_split
    has_deps = True
except:
    has_deps = False

def _raise_dep_error():
    raise Exception('This method requires pandas, numpy, scikit-learn. Install it using pip install pandas numpy scikit-learn')

def add_datepart(df, col_name, drop=True, errors="raise"):
    '''
    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    df                  Required DataFrame.
    col_name            Required String.
    drop                Optional Boolean. If True drop the original col_name.
    errors              Optional
    ===============     ====================================================================

    :return: None
    '''
    if not has_deps:
        _raise_dep_error()
    col = df[col_name]
    col_dtype = col.dtype
    if isinstance(col_dtype, pd.core.dtypes.dtypes.DatetimeTZDtype):
        col_dtype = np.datetime64
    if not np.issubdtype(col_dtype, np.datetime64):
        df[col_name] = col = pd.to_datetime(col, infer_datetime_format=True, errors=errors)
    attr = ['Year', 'Month', 'Week', 'Day', 'Dayofweek', 'Dayofyear',
            'Is_month_end', 'Is_month_start', 'Is_quarter_end', 'Is_quarter_start', 'Is_year_end', 'Is_year_start']
    if df[col_name][0].time(): attr = attr + ['Hour', 'Minute', 'Second']
    for n in attr: df[col_name + n] = getattr(col.dt, n.lower())
    if drop: df.drop(col_name, axis=1, inplace=True)

def _scale(df, mapper=None):
    '''
    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    df                  DataFrame to be scaled.
    mapper              Parameters used for scaling.
    ===============     ====================================================================

    :return: mapper if passed as None
    '''
    if mapper is None:
        map_f = [([n], StandardScaler()) for n in df.columns if is_numeric_dtype(df[n])]
        mapper = DataFrameMapper(map_f).fit(df)
    df[mapper.transformed_names_] = mapper.transform(df)
    return mapper

def process_df(df, target=None, do_scale=False, add_date_feats=False, mapper=None, test_sz=0.2):
    '''
    This function preprocess the dataframe in following order :
    a. drops SHAPE column,
    b. creates target feature,
    c. fill missing values,
    d. scales the numerical features,
    e. one hot encode categorical features,
    f. train-test split,
    and returns the preprocessed dataframe, target variable, and mapper.
    ==============   ====================================================================
    **Argument**        **Description**
    --------------   --------------------------------------------------------------------
    df               DataFrame to be preprocessed.
    --------------   --------------------------------------------------------------------
    y                Optional string. Response(or target) variable.
    --------------   --------------------------------------------------------------------
    do_scale         Optional Boolean. If True scales the numerical features else do not.
    --------------   --------------------------------------------------------------------
    mapper           Optional.  Works if do_scale is True, it contains the
                     parameters(mean and standard deviation) obtained from training set
                     and can be used for  test set.
    --------------   --------------------------------------------------------------------

    :return: training set features, test set features, train set targets, test set targets, np.ndarray, DataFrameMapper
    '''
    if not has_deps:
        _raise_dep_error()

    if target is None :
        print("none")
        raise Exception('y(target variable) not found!!!!!')
    df = df.copy()
    for i in df:
        if str(df[i].dtype)=='geometry':
            shape = df.pop(i)
    if target is not None:
        if not is_numeric_dtype(df[target]):
            type_y = 'cat'
            y= df.pop(target).astype('category').cat.codes
        else:
            type_y = 'num'
            y = df.pop(target).values
    else:
        type_y = None
        y = None
    if df.isnull().sum().sum():
        df.fillna(df.median(), inplace=True)#for numerical columns!!
        df.fillna(df.mode().iloc[0,:], inplace=True)#for remaining categorical columns
    if add_date_feats:
        for i in df:
            if np.issubdtype(df[i].dtype, np.datetime64):
                add_datepart(df, i)
    if do_scale:
        mapper = _scale(df, mapper)
    cat_cols = df.select_dtypes(include=['object','category']).columns
    for i in cat_cols:
        if df[i].nunique()>50:
            df.drop(i, inplace=True, axis=1)
            logging.warning('feature {i} contains more than 50 categories, dropping!!'.format(i=i))
    df = pd.get_dummies(df)
    if (type_y=='num'): #regression type
        X_train, X_test, y_train, y_test = train_test_split(df, y, test_size=0.2, random_state=42)
        return X_train, X_test, y_train, y_test, mapper
    elif (type_y=='cat') : #when y is not None and category type.
        split = StratifiedShuffleSplit(n_splits=1, test_size=0.2, random_state=42)
        for train_index, test_index in split.split(df, y):
            X_train = df.loc[train_index]
            y_train = y[train_index]
            X_test = df.loc[test_index]
            y_test = y[test_index]
        return X_train,  X_test, y_train, y_test, mapper



from fastai.vision import ImageSegment, Image
from fastai.vision.image import open_image, show_image, pil2tensor
from fastai.vision.data import SegmentationProcessor, ImageList
from fastai.layers import CrossEntropyFlat
from fastai.basic_train import LearnerCallback
import torch
import warnings
import PIL
import numpy as np
import os
import math

def _class_array_to_rbg(ca : 'classified_array', cm : 'color_mapping', nodata=0):
    im = np.expand_dims(ca, axis=2).repeat(3, axis=2)
    white_mask = im[im == nodata]
    for x in (np.unique(im)):
        if not x == nodata:
            for i in range(3):
                im[:,:,i][im[:,:,i] == x] = cm[x][i]
    im[white_mask] = 255
    return im

#def _show_batch_unet_multispectral(self, nrows=3, ncols=3, n_items=None, index=0, rgb_bands=None, nodata=0, alpha=0.7, imsize=5): # Proposed Parameters 
def _show_batch_unet_multispectral(self, rows=3, alpha=0.7, **kwargs): # parameters adjusted in kwargs
    import matplotlib.pyplot as plt
    from .._data import _tensor_scaler
   
    nrows = rows
    ncols = 3
    if kwargs.get('ncols', None) is not None:
        ncols = kwargs.get('ncols')
    
    n_items = None
    if kwargs.get('n_items', None) is not None:
        n_items = kwargs.get('n_items')

    ds = self.train_ds
    if kwargs.get('type_ds', None) is not None:
        type_ds = kwargs.get('type_ds')
        if getattr(self, type_ds, None) is not None:
            ds = getattr(self, type_ds)
        else:
            e = Exception(f'could not find {str(type_ds)} in data.')
            raise(e)

    rgb_bands = self._symbology_rgb_bands
    if kwargs.get('rgb_bands', None) is not None:
        rgb_bands = kwargs.get('rgb_bands')

    nodata = 0
    if kwargs.get('nodata', None) is not None:
        nodata = kwargs.get('nodata')

    index = 0
    if kwargs.get('index', None) is not None:
        index = kwargs.get('index')

    imsize = 5
    if kwargs.get('imsize', None) is not None:
        imsize = kwargs.get('imsize')


    e = Exception('`rgb_bands` should be a valid band_order, list or tuple of length 3 or 1.')
    symbology_bands = []
    if not ( len(rgb_bands) == 3 or len(rgb_bands) == 1 ):
        raise(e)
    for b in rgb_bands:
        if type(b) == str:
            b_index = self._bands.index(b)
        elif type(b) == int:
            self._bands[b] # To check if the band index specified by the user really exists.
            b_index = b
        else:
            raise(e)
        symbology_bands.append(b_index)

    # Get Batch
    if n_items is None:
        n_items = nrows*ncols
    else:
        nrows = math.ceil(n_items/ncols)
    n_items = min(n_items, len(self.x))

    x_batch, y_batch = [], []
    for i in range(index, index+n_items):
        x_batch.append(ds.x[i].data)
        y_batch.append(ds.y[i].data[0])
    x_batch = self._min_max_scaler(torch.stack(x_batch))
    symbology_x_batch = x_batch[:, symbology_bands].cpu().numpy()

    # Channel first to channel last for plotting
    symbology_x_batch = np.rollaxis(symbology_x_batch, 1, 4)
    if symbology_x_batch.max() < 1.5:
        symbology_x_batch = symbology_x_batch.clip(0, 1)

    # Size for plotting
    fig, ax = plt.subplots(nrows=nrows, ncols=ncols, figsize=(ncols*imsize, nrows*imsize))
    idx = 0
    for r in range(nrows):
        for c in range(ncols):
            if idx < symbology_x_batch.shape[0]:
                axi  = ax[r][c]
                axi.imshow(symbology_x_batch[idx])
                y_rgb = _class_array_to_rbg(y_batch[idx], self._multispectral_color_mapping, nodata)
                axi.imshow(y_rgb, alpha=alpha)
                axi.axis('off')
            else:
                ax[r][c].axis('off')
            idx+=1

class ArcGISImageSegment(Image):
    "Support applying transforms to segmentation masks data in `px`."
    def __init__(self, x, cmap=None, norm=None):
        super(ArcGISImageSegment, self).__init__(x)
        self.cmap = cmap
        self.mplnorm = norm

    def lighting(self, func, *args, **kwargs):
        return self

    def refresh(self):
        self.sample_kwargs['mode'] = 'nearest'
        return super().refresh()

    @property
    def data(self):
        "Return this image pixels as a `LongTensor`."
        return self.px.long()

    def show(self, ax=None, figsize:tuple=(3,3), title=None, hide_axis:bool=True,
        cmap='tab20', alpha:float=0.5, **kwargs):
        "Show the `ImageSegment` on `ax`."
        ax = show_image(self, ax=ax, hide_axis=hide_axis, cmap=self.cmap, figsize=figsize,
                        interpolation='nearest', alpha=alpha, vmin=0, norm=self.mplnorm, **kwargs)
        if title: ax.set_title(title)

class ArcGISMultispectralImageSegment():
    def __init__(self, tensor):
        self.data = tensor
        self.size = tensor.shape

def is_no_color(color_mapping):
    if isinstance(color_mapping, dict):
        color_mapping = list(color_mapping.values())
    return (np.array(color_mapping) == [-1., -1., -1.]).any()

class ArcGISSegmentationLabelList(ImageList):
    "`ItemList` for segmentation masks."
    _processor = SegmentationProcessor
    def __init__(self, items, classes=None, class_mapping=None, color_mapping=None, **kwargs):
        super().__init__(items, **kwargs)
        self.class_mapping = class_mapping
        self.color_mapping = color_mapping
        self.copy_new.append('classes')
        self.classes, self.loss_func = classes, CrossEntropyFlat(axis=1)

        if is_no_color(list(color_mapping.values())):
            self.cmap = 'tab20'  ## compute cmap from palette
            import matplotlib as mpl
            bounds = list(color_mapping.keys())
            if len(bounds) < 3: # Two handle two classes i am adding one number to the classes which is not already in bounds
                bounds = bounds + [max(bounds)+1]
            self.mplnorm = mpl.colors.BoundaryNorm(bounds, len(bounds))
        else:
            import matplotlib as mpl
            bounds = list(color_mapping.keys())
            if len(bounds) < 3: # Two handle two classes i am adding one number to the classes which is not already in bounds
                bounds = bounds + [max(bounds)+1]
            self.cmap = mpl.colors.ListedColormap(np.array(list(color_mapping.values()))/255)
            self.mplnorm = mpl.colors.BoundaryNorm(bounds, self.cmap.N)

        if len(color_mapping.keys()) == 1:
            self.cmap = 'tab20'
            self.mplnorm = None
        

    def open(self, fn):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning) # EXIF warning from TiffPlugin
            x = PIL.Image.open(fn)
            if x.palette is not None:
                x = x.convert('P')
            else:
                x = x.convert('L')
            x = pil2tensor(x, np.float32)

        return ArcGISImageSegment(x, cmap=self.cmap, norm=self.mplnorm)

    def analyze_pred(self, pred, thresh:float=0.5): 
        label_mapping = {(idx + 1):value for idx, value in enumerate(self.class_mapping.keys())}
        out = pred.argmax(dim=0)[None]
        predictions = torch.zeros_like(out)
        for key, value in label_mapping.items():
            predictions[out==key] = value
        return predictions

    def reconstruct(self, t): 
        return ArcGISImageSegment(t, cmap=self.cmap, norm=self.mplnorm)

class ArcGISSegmentationItemList(ImageList):
    "`ItemList` suitable for segmentation tasks."
    _label_cls, _square_show_res = ArcGISSegmentationLabelList, False

class ArcGISSegmentationMSLabelList(ArcGISSegmentationLabelList):
    def open(self, fn):
        import gdal
        path = str(os.path.abspath(fn))
        x = gdal.Open(path).ReadAsArray()
        x = torch.tensor(x.astype(np.long))[None]
        return ArcGISImageSegment(x, cmap=self.cmap, norm=self.mplnorm)

class ArcGISSegmentationMSItemList(ImageList):
    "`ItemList` suitable for segmentation tasks."
    _label_cls, _square_show_res = ArcGISSegmentationMSLabelList, False
    def open(self, fn):
        import gdal
        path = str(os.path.abspath(fn))
        x = gdal.Open(path).ReadAsArray()
        #x = ArcGISImageMSSegment(x.astype(np.float32))
        x = torch.tensor(x.astype(np.float32))
        #x = ArcGISImageSegment( x[:3, ] )
        x = ArcGISMultispectralImageSegment(x)
        return x

class LabelCallback(LearnerCallback):
    def __init__(self, learn):
        super().__init__(learn)
        import pdb
        self.label_mapping = {value:(idx+1) for idx, value in enumerate(learn.data.class_mapping.keys())}
        
    def on_batch_begin(self, last_input, last_target, **kwargs):
        modified_target = torch.zeros_like(last_target)
        for label, idx in self.label_mapping.items():
            modified_target[last_target==label] = idx
        return {'last_input':last_input, 'last_target':modified_target}